#!/usr/bin/env python
# -*- coding: utf-8 -*-


class SolverException(Exception):
    pass


class UnexpectedlySatisfiable(SolverException):
    pass


class InvalidConstraint(SolverException):
    pass


class InvalidDependencyString(InvalidConstraint):
    pass


class NoPackageFound(SolverException):
    def __init__(self, requirement, *a, **kw):
        # NOTE: Working around a circular import
        from simplesat.constraints.requirement import Requirement

        if not isinstance(requirement, Requirement):
            msg = "NoPackageFound expects a Requirement as first arg. Got {!r}"
            raise TypeError(msg.format(requirement))
        super(NoPackageFound, self).__init__(*a, **kw)
        self.requirement = requirement
        self.args = self.args or (str(requirement),)


class MissingInstallRequires(NoPackageFound):
    pass


class MissingConflicts(NoPackageFound):
    pass


class SatisfiabilityError(SolverException):
    def __init__(self, unsat):
        self.unsat = unsat
        super(SatisfiabilityError, self).__init__(self.reason)

    @property
    def reason(self):
        return self.unsat.to_string()


class SatisfiabilityErrorWithHint(SatisfiabilityError):
    """ A satistibiality error class with information about minimally
    unsatisfiable problem.

    This is used when one wants to give more human-readable error messages
    about conflicts and other satistiability issues.
    """
    def __init__(self, unsat, conflicting_jobs):
        self.conflicting_jobs = conflicting_jobs
        super(SatisfiabilityErrorWithHint, self).__init__(unsat)

    @property
    def hint_pretty_string(self):
        return (
            u"The following jobs are conflicting:\n{}".format(
                u"\n".join("    {}".format(str(job)) for job in self.conflicting_jobs)
            )
        )



import unittest

from okonomiyaki.versions import EnpkgVersion

from ..kinds import (
    Any, EnpkgUpstreamMatch, Equal, GEQ, GT, LEQ, LT, Not, IVersionConstraint
)
from ..multi import MultiConstraints


V = EnpkgVersion.from_string


class TestConstraintMisc(unittest.TestCase):
    def test_ensure_can_compare(self):
        # Given
        left = Equal(V("1.2-1"))

        class DummyVersion(object):
            pass
        right = Equal(DummyVersion())

        # When/Then
        with self.assertRaises(TypeError):
            self.assertTrue(left.matches(right))

    def test_hashing(self):
        # Given
        v1 = V("1.2.1-1")
        v2 = V("1.2.1-1")

        # When/Then
        self.assertEqual(hash(v1), hash(v2))
        self.assertEqual(hash(GEQ(v1)), hash(GEQ(v2)))

    def test_comparison(self):
        # Given
        class DummyVersion(IVersionConstraint):
            def matches(self, version):
                return True

        v1 = V("1.2.1-1")
        v2 = V("1.2.1-1")

        # Then
        self.assertEqual(DummyVersion(v1), DummyVersion(v2))
        self.assertFalse(DummyVersion(v1) != DummyVersion(v2))


class TestAreCompatible(unittest.TestCase):
    def test_simple_equal(self):
        # Given
        constraint = Equal(V("1.2-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = Equal(V("1.2-1"))
        candidate = V("1.3-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

    def test_simple_any(self):
        # Given
        constraint = Any()
        candidate = V("1.2.3-1")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

    def test_simple_leq(self):
        # Given
        constraint = LEQ(V("1.2-2"))
        candidate = V("1.2-1")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = LEQ(V("1.2-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = LEQ(V("1.1-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

    def test_simple_lt(self):
        # Given
        constraint = LT(V("1.2-2"))
        candidate = V("1.2-1")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = LT(V("1.2-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

        # Given
        constraint = LT(V("1.1-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

    def test_simple_geq(self):
        # Given
        constraint = GEQ(V("1.2-2"))
        candidate = V("1.2-3")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = GEQ(V("1.2-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = GEQ(V("1.2-2"))
        candidate = V("1.2-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

    def test_simple_gt(self):
        # Given
        constraint = GT(V("1.2-2"))
        candidate = V("1.2-3")

        # When/Then
        self.assertTrue(constraint.matches(candidate))

        # Given
        constraint = GT(V("1.2-1"))
        candidate = V("1.2-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

        # Given
        constraint = GT(V("1.2-2"))
        candidate = V("1.2-1")

        # When/Then
        self.assertFalse(constraint.matches(candidate))

    def test_simple_not(self):
        # Given
        constraint = Not(V("1.2-1"))

        # When/Then
        self.assertTrue(constraint.matches(V("1.2-3")))
        self.assertTrue(constraint.matches(V("1.2-2")))
        self.assertTrue(constraint.matches(V("1.2-6")))
        self.assertFalse(constraint.matches(V("1.2-1")))

    def test_simple_enpkg_upstream_match(self):
        # Given
        constraint = EnpkgUpstreamMatch(V("1.2"))

        # When/Then
        self.assertTrue(constraint.matches(V("1.2-3")))
        self.assertTrue(constraint.matches(V("1.2-1")))
        self.assertTrue(constraint.matches(V("1.2-6")))
        self.assertFalse(constraint.matches(V("1.2.1-1")))
        self.assertFalse(constraint.matches(V("1.3-2")))
        self.assertFalse(constraint.matches(V("2.0-3")))


class TestMultiConstraints(unittest.TestCase):
    def test_simple(self):
        # Given
        constraints_string = ">= 1.3, < 2.0"

        # When
        constraints = MultiConstraints._from_string(constraints_string)

        # Then
        self.assertFalse(constraints.matches(V("1.2-3")))
        self.assertTrue(constraints.matches(V("1.3.0")))
        self.assertTrue(constraints.matches(V("1.3.0-1")))
        self.assertTrue(constraints.matches(V("1.3.0-3")))
        self.assertTrue(constraints.matches(V("1.4.0-1")))
        self.assertFalse(constraints.matches(V("2.0-1")))

    def test_comparison(self):
        # Given
        versions = (
            V("1.2-3"),
            V("1.3.0"),
            V("1.3.0-1"),
            V("1.3.0-3"),
            V("1.4.0-1"),
            V("2.0-1")
        )
        constraints_string1 = ">= 1.3, < 2.0"
        constraints_string2 = "< 2.0, >= 1.3"

        # When
        constraints1 = MultiConstraints._from_string(constraints_string1)
        constraints2 = MultiConstraints._from_string(constraints_string2)

        # Then
        for version in versions:
            self.assertEqual(
                constraints1.matches(version), constraints2.matches(version)
            )



import collections
import re

import six

from simplesat.constraints.kinds import (
    Any, EnpkgUpstreamMatch, Equal, GEQ, GT, LEQ, LT, Not
)

from simplesat.errors import InvalidConstraint


# NOTE: The _DISTRIBUTION_R regex is based on PEP508. Additionally, we remove
# the '-' because it breaks too many other things to do otherwise. We must also
# permit a leading or trailling underscore because of names like
# `_distribute_remove` re.Scanner seems to ignore (?i), so we have to write
# case-insensitivity into the regex manually.
_DISTRIBUTION_NAME_R = r"(?!\.)(?:\.?\w+)+(?<!\.)"
_DISTRIBUTION_R = "({})".format(_DISTRIBUTION_NAME_R)
_VERSION_R = r"((?=\d){}(?:-\w+)?$)".format(_DISTRIBUTION_NAME_R)
_EQUAL_R = r"=="
_GEQ_R = r">="
_GT_R = r">"
_LEQ_R = r"<="
_LT_R = r"<"
_NOT_R = r"!="
_ENPKG_UPSTREAM_MATCH_R = r"\^="
_ANY_R = r"\*"
_WS_R = r" +"

_CONSTRAINTS_SCANNER = re.Scanner([
    (_VERSION_R, lambda scanner, token: VersionToken(token)),
    (_EQUAL_R, lambda scanner, token: EqualToken(token)),
    (_GEQ_R, lambda scanner, token: GEQToken(token)),
    (_GT_R, lambda scanner, token: GTToken(token)),
    (_LEQ_R, lambda scanner, token: LEQToken(token)),
    (_LT_R, lambda scanner, token: LTToken(token)),
    (_NOT_R, lambda scanner, token: NotToken(token)),
    (_ENPKG_UPSTREAM_MATCH_R,
        lambda scanner, token: EnpkgUpstreamMatchToken(token)),
    (_ANY_R, lambda scanner, token: AnyToken(token)),
    (_WS_R, lambda scanner, token: None),
])

_REQUIREMENTS_SCANNER = re.Scanner([
    (_VERSION_R, lambda scanner, token: VersionToken(token)),
    (_DISTRIBUTION_R, lambda scanner, token: DistributionNameToken(token)),
    (_EQUAL_R, lambda scanner, token: EqualToken(token)),
    (_GEQ_R, lambda scanner, token: GEQToken(token)),
    (_GT_R, lambda scanner, token: GTToken(token)),
    (_LEQ_R, lambda scanner, token: LEQToken(token)),
    (_LT_R, lambda scanner, token: LTToken(token)),
    (_NOT_R, lambda scanner, token: NotToken(token)),
    (_ENPKG_UPSTREAM_MATCH_R,
        lambda scanner, token: EnpkgUpstreamMatchToken(token)),
    (_ANY_R, lambda scanner, token: AnyToken(token)),
    (_WS_R, lambda scanner, token: None),
])


class Token(object):
    kind = None

    def __init__(self, value=None):
        self.value = value


class CommaToken(Token):
    kind = "comma"


class DistributionNameToken(Token):
    kind = "distribution_name"


class AnyToken(Token):
    kind = "any"


class VersionToken(Token):
    kind = "version"


class ComparisonToken(Token):
    kind = "comparison"


class LEQToken(ComparisonToken):
    kind = "leq"


class LTToken(ComparisonToken):
    kind = "lt"


class GEQToken(ComparisonToken):
    kind = "geq"


class GTToken(ComparisonToken):
    kind = "gt"


class EnpkgUpstreamMatchToken(ComparisonToken):
    kind = "enpkg_upstream"


class EqualToken(ComparisonToken):
    kind = "equal"


class NotToken(ComparisonToken):
    kind = "not"


_OPERATOR_TO_SPEC = {
    EnpkgUpstreamMatchToken: EnpkgUpstreamMatch,
    EqualToken: Equal,
    GEQToken: GEQ,
    GTToken: GT,
    LEQToken: LEQ,
    LTToken: LT,
    NotToken: Not,
}


def _spec_factory(comparison_token):
    klass = _OPERATOR_TO_SPEC.get(comparison_token.__class__, None)
    if klass is None:
        msg = "Unsupported comparison token {0!r}".format(comparison_token)
        raise InvalidConstraint(msg)
    else:
        return klass


def _tokenize(scanner, requirement_string):
    requirement_string = requirement_string.strip()

    if len(requirement_string) == 0:
        return [[AnyToken()]]
    else:
        tokens = []

        for part in requirement_string.split(","):
            scanned, remaining = scanner.scan(part.strip())
            if len(remaining) > 0:
                msg = "{0!r}"
                for tok in scanned:
                    if isinstance(tok, DistributionNameToken):
                        msg += "(distribution name: {0!r})".format(tok.value)
                    if isinstance(tok, VersionToken):
                        msg += "(version: {0!r})".format(tok.value)
                msg += "(unparsed {0!r})".format(remaining)
                raise InvalidConstraint(msg.format(requirement_string))
            elif len(scanned) > 0:
                tokens.append(scanned)
        return tokens


def _operator_factory(operator, version, version_factory):
    operator = _spec_factory(operator)
    version = version_factory(version.value)
    return operator(version)


class _RawConstraintsParser(object):
    """A simple parser for requirement strings."""
    def __init__(self):
        self._scanner = _CONSTRAINTS_SCANNER

    def parse(self, requirement_string, version_factory):
        def compute_constraint(requirement_block):
            if len(requirement_block) == 2:
                operator, version = requirement_block
                return _operator_factory(operator, version, version_factory)
            elif len(requirement_block) == 1:
                assert isinstance(requirement_block[0], AnyToken)
                return Any()
            else:
                msg = "{0!r}".format(requirement_string)
                raise InvalidConstraint(msg)

        constraints = []
        tokens_blocks = _tokenize(self._scanner, requirement_string)

        for requirement_block in tokens_blocks:
            constraints.append(compute_constraint(requirement_block))

        return tuple(constraints)


class _RawRequirementParser(object):
    """A simple parser for requirement strings."""
    def __init__(self):
        self._scanner = _REQUIREMENTS_SCANNER

    def parse(self, requirement_string, version_factory):
        def compute_constraint(requirement_block):
            msg = "{0!r}".format(requirement_string)

            if len(requirement_block) == 3:
                distribution, operator, version = requirement_block
                if not isinstance(distribution, DistributionNameToken):
                    raise InvalidConstraint(msg + ' (bad distribution name)')
                if not isinstance(version, VersionToken):
                    raise InvalidConstraint(msg + ' (bad version)')
                name = distribution.value
                op = _operator_factory(operator, version, version_factory)
                return name, (op,)
            elif len(requirement_block) == 2:
                distribution, operator = requirement_block
                name = distribution.value
                if isinstance(operator, AnyToken):
                    return name, (Any(),)
                else:
                    raise InvalidConstraint(msg)
            elif len(requirement_block) == 1:
                name = requirement_block[0].value
                return name, (Any(),)
            else:
                raise InvalidConstraint(msg)

        named_constraints = collections.defaultdict(list)
        tokens_blocks = _tokenize(self._scanner, requirement_string)

        for requirement_block in tokens_blocks:
            name, constraint = compute_constraint(requirement_block)
            named_constraints[name].extend(constraint)

        return dict(
            (name, tuple(constraints))
            for name, constraints in six.iteritems(named_constraints)
        )



import textwrap
import unittest

from okonomiyaki.versions import EnpkgVersion

from simplesat.constraints import PrettyPackageStringParser
from simplesat.errors import NoPackageFound
from simplesat.package import RepositoryInfo, RepositoryPackageMetadata
from simplesat.repository import Repository


V = EnpkgVersion.from_string


class TestRepository(unittest.TestCase):
    def packages_from_definition(self, packages_definition,
                                 repository_info=None):
        repository_info = repository_info or RepositoryInfo("dummy")
        parser = PrettyPackageStringParser(EnpkgVersion.from_string)

        return [
            RepositoryPackageMetadata(
                parser.parse_to_package(line), repository_info
            )
            for line in packages_definition.splitlines()
        ]

    def test_simple(self):
        # Given
        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(packages_definition)

        # When
        repository = Repository()

        # Then
        self.assertEqual(len(repository), 0)

        # When
        repository = Repository(packages)

        # Then
        self.assertEqual(len(repository), len(packages))
        self.assertEqual(list(repository), packages)

        # When
        repository = Repository(reversed(packages))

        # Then
        self.assertEqual(len(repository), len(packages))
        self.assertEqual(list(repository), packages)

    def test_update(self):
        # Given
        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(packages_definition)

        # When
        repository = Repository(packages[:4])

        # Then
        self.assertEqual(len(repository), 4)

        # When
        repository.update(packages[4:])

        # Then
        self.assertEqual(len(repository), len(packages))
        self.assertEqual(list(repository), packages)

    def test_find_package(self):
        # Given
        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(packages_definition)
        repository = Repository(packages)

        # When
        package = repository.find_package("nose", V("1.3.0-1"))

        # Then
        self.assertEqual(package.name, "nose")
        self.assertEqual(package.version, V("1.3.0-1"))

    def test_find_unavailable_package(self):
        # Given
        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(packages_definition)
        repository = Repository(packages)

        # When/Then
        with self.assertRaises(NoPackageFound):
            repository.find_package("nose", V("1.4.0-1"))

        repository.find_package("nose", V("1.3.0-1"))
        with self.assertRaises(NoPackageFound):
            repository.find_package("nono", V("1.3.0-1"))

    def test_find_packages(self):
        # Given
        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.3.0-1
        nose 1.2.1-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(packages_definition)
        repository = Repository(packages)

        r_versions = (V("1.2.1-1"), V("1.3.0-1"), V("1.3.0-2"))

        # When
        packages = repository.find_packages("nose")

        # Then
        self.assertEqual(len(packages), 3)
        self.assertEqual(tuple(p.version for p in packages), r_versions)

        # When
        packages = repository.find_packages("non_existing_package")

        # Then
        self.assertEqual(len(packages), 0)

    def test_contains(self):
        # Given
        repository_info1 = RepositoryInfo("repo1")
        repository_info2 = RepositoryInfo("repo2")

        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(
            packages_definition, repository_info1
        )
        repository = Repository(packages)

        # When/Then
        for package in repository:
            self.assertTrue(package in repository)

        # Given
        package = RepositoryPackageMetadata._from_pretty_string(
            "dummy 1.0.1-1", RepositoryInfo("another")
        )

        # When/Then
        self.assertFalse(package in repository)

        # Given
        dummy_repository1 = self.packages_from_definition(
            "dummy 1.0.1-1", repository_info1
        )[0]

        dummy_repository2 = self.packages_from_definition(
            "dummy 1.0.1-1", repository_info2
        )[0]

        # When/Then
        self.assertTrue(dummy_repository1 in repository)
        self.assertFalse(dummy_repository2 in repository)

    def test_add_package(self):
        # Given
        repository_info1 = RepositoryInfo("repo1")

        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(
            packages_definition, repository_info1
        )
        repository = Repository()

        # When
        repository.add_package(packages[0])

        # Then
        self.assertTrue(packages[0] in repository)
        self.assertEqual(len(repository), 1)
        self.assertEqual(list(repository), packages[:1])

    def test_regression_185(self):
        # Given
        repository_info1 = RepositoryInfo("repo1")

        packages_definition = textwrap.dedent(u"""\
        dummy 1.0.1-1
        dummy_with_appinst 1.0.0-1
        dummy_with_entry_points 1.0.0-1
        dummy_with_proxy 1.3.40-3
        dummy_with_proxy_scripts 1.0.0-1
        dummy_with_proxy_softlink 1.0.0-1
        nose 1.2.1-1
        nose 1.3.0-1
        nose 1.3.0-2\
        """)
        packages = self.packages_from_definition(
            packages_definition, repository_info1
        )
        repository = Repository()

        package = packages[0]

        # When
        repository.find_packages(package.name)
        repository.add_package(package)

        # Then
        self.assertTrue(packages[0] in repository)
        self.assertEqual(len(repository), 1)
        self.assertEqual(list(repository), [package])



from simplesat.errors import SatisfiabilityError  # noqa
from .minisat import MiniSATSolver  # noqa


def is_satisfiable(rules):
    s = MiniSATSolver(rules)
    try:
        s.search()
        return True
    except SatisfiabilityError:
        return False



import enum
from collections import OrderedDict, deque
import logging

from .constraints import ConflictRequirement, InstallRequirement
from .errors import (
    MissingConflicts, MissingInstallRequires, NoPackageFound, SolverException
)
from .request import JobType


logger = logging.getLogger(__name__)

INDENT = 4


class RuleType(enum.Enum):
    internal_allow_update = 1
    job_install = 2
    job_remove = 3
    job_update = 4
    package_requires = 7
    package_conflicts = 8
    package_same_name = 10
    package_installed = 12
    package_broken = 100

    internal = 256

    @property
    def is_job(self):
        return self in (
            RuleType.job_install,
            RuleType.job_remove,
            RuleType.job_update
        )


class PackageRule(object):
    @classmethod
    def _from_string(cls, rule_string, pool):
        """
        Creates a PackageRule from a rule string, e.g.
            '-numpy-1.6.0 | numpy-1.7.0'

        Because package full name -> id is not 1-to-1 mapping, this may fail
        when a package has multiple ids. This is mostly used for testing, to
        write reference rules a bit more easily.
        """
        packages_string = (s.strip() for s in rule_string.split("|"))
        package_literals = []
        for package_string in packages_string:
            if package_string.startswith("-"):
                positive = False
                package_string = package_string[1:]
            else:
                positive = True

            requirement = InstallRequirement.from_package_string(
                package_string)
            package_candidates = pool.what_provides(requirement)
            if len(package_candidates) == 0:
                msg = "No candidate for package {0!r}".format(package_string)
                raise NoPackageFound(requirement, msg)
            elif len(package_candidates) > 1:
                msg = "> 1 candidate for package {0!r} requirement, cannot " \
                      "create rule from it" % package_string
                raise SolverException(msg)
            else:
                _id = pool.package_id(package_candidates[0])
                if positive:
                    package_literals.append(_id)
                else:
                    package_literals.append(-_id)
        return cls(package_literals, None, requirements=(requirement,))

    def __init__(self, literals, reason, requirements=None):
        self.literals = tuple(sorted(literals))
        self._reason = RuleType(reason)
        assert isinstance(requirements, (tuple, type(None)))
        self._requirements = requirements or ()

    @property
    def is_assertion(self):
        return len(self.literals) == 1

    @property
    def reason(self):
        return self._reason

    def _pretty_literals(self, pool, literals, sign=True, unique=False):
        parts = (pool.id_to_string(literal) for literal in literals)
        if not sign:
            parts = (p[1:] for p in parts)
        if unique:
            parts = OrderedDict.fromkeys(parts).keys()
        return " | ".join(parts)

    def to_string(self, pool, unique=False):
        s = self._pretty_literals(pool, self.literals, unique=unique)

        if self._reason == RuleType.job_install:
            rule_desc = "Install command rule ({})".format(s)
        elif self._reason == RuleType.job_update:
            rule_desc = "Update to latest command rule ({})".format(s)
        elif self._reason == RuleType.job_remove:
            rule_desc = "Remove command rule ({})".format(s)
        elif self._reason == RuleType.package_same_name:
            s = self._pretty_literals(
                pool, (abs(i) for i in self.literals), unique=unique)
            rule_desc = "Can only install one of: ({})".format(s)
        elif self._reason == RuleType.package_installed:
            parts = [pool.id_to_string(abs(literal))
                     for literal in self.literals]
            s = " | ".join(parts)
            rule_desc = "Should install one of: ({})".format(s)
        elif self._reason == RuleType.package_requires:
            source_ids = [abs(self.literals[0])]
            source = self._pretty_literals(pool, source_ids, unique=unique)
            source = source[1:]  # trim off +/- sign
            s = self._pretty_literals(pool, self.literals[1:], unique=unique)
            rule_desc = "{} requires ({})".format(source, s)
        elif self._reason == RuleType.package_conflicts:
            left_ids, right_ids = self.literals[0:1], self.literals[1:]
            left = self._pretty_literals(pool, left_ids, sign=False)
            right = self._pretty_literals(pool, (abs(i) for i in right_ids))
            rule_desc = "{} conflicts with {}".format(left, right)
        elif self._reason == RuleType.package_broken:
            package_id = self.literals[0]
            # Trim the sign
            package_str = pool.id_to_string(abs(package_id))
            msg = "{} was ignored because it depends on missing packages"
            rule_desc = msg.format(package_str)
        else:
            rule_desc = s

        if self._requirements:
            reqs = ' <- '.join("'{}'".format(r) for r in self._requirements)
            rule_desc = "Requirements: {reqs}\n{indent}{rule}".format(
                reqs=reqs, rule=rule_desc, indent=" " * INDENT)

        return rule_desc

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.literals == other.literals)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.literals)


class RulesGenerator(object):
    def __init__(self, pool, request,
                 installed_package_ids=None, strict=False):
        self._rules_set = OrderedDict()
        self._pool = pool

        self.request = request
        self.installed_package_ids = installed_package_ids or OrderedDict()
        self.added_package_ids = set()
        self.strict = strict

    def iter_rules(self):
        """
        Return an iterator over each created rule.
        """
        self.added_package_ids = set()
        # This attaches the job requirement to the created rule. We need
        # to run it first because duplicated rules are ignored. Otherwise,
        # we'll end up keeping the rule instance that doesn't know it should be
        # associated with a job.
        self._add_job_rules()
        for package in self.installed_package_ids.values():
            self._add_installed_package_rules(package)
            self._add_package_rules(package)
        return self._rules_set

    # ------------------------------
    # API to create individual rules
    # ------------------------------
    def _create_dependency_rule(self, package, install_requires, reason,
                                requirements=None):
        """
        Create the rule for the install_requires of a package.

        This dependency is of the form (-A | R1 | R2 | R3) where R* are
        the set of packages provided by the dependency requirement.

        Parameters
        ----------
        package: PackageMetadata
            The package with a requirement
        install_requires: sequence
            Sequence of packages that fulfill the requirement.
        reason: RuleType
            A valid PackageRule.reason value
        requirements: tuple of Requirement
            Optional requirements explaining the rule's origin.

        Returns
        -------
        rule: PackageRule or None
        """
        literals = [-self._pool.package_id(package)]

        for dependency in install_requires:
            if dependency != package:
                literals.append(self._pool.package_id(dependency))

        return PackageRule(literals, reason, requirements=requirements)

    def _create_conflicts_rule(self, issuer, provider,
                               reason, requirements=None):
        """
        Return a conflict rule between issuer and provider or None if issuer ==
        provider.

        The rule is of the form (-A | -B)

        Parameters
        ----------
        issuer: PackageMetadata
            Package declaring the conflict
        provider: PackageMetadata
            Package causing the conflict
        reason: RuleType
            One of PackageRule.reason
        requirements: tuple of Requirement
            Optional requirements explaining the rule's origin.

        Returns
        -------
        rule: PackageRule or None
        """
        if issuer != provider:
            return PackageRule([-self._pool.package_id(issuer),
                                -self._pool.package_id(provider)],
                               reason, requirements=requirements)
        else:
            return None

    def _create_install_one_of_rule(self, packages, reason, requirements=None):
        """
        Creates a rule to Install one of the given packages.

        The rule is of the form (A | B | C)

        Parameters
        ----------
        packages: sequence
            List of packages to choose from
        reason: RuleType
            One of PackageRule.reason
        requirements: tuple of Requirement
            Optional requirements explaining the rule's origin.

        Returns
        -------
        rule: PackageRule
        """
        literals = [self._pool.package_id(p) for p in packages]
        return PackageRule(literals, reason, requirements=requirements)

    def _create_remove_rule(self, package, reason, requirements=None):
        """
        Create the rule to remove a package.

        For a package A, the rule is simply (-A)

        Parameters
        ----------
        package: PackageMetadata
            The package with a requirement
        reason: RuleType
            One of PackageRule.reason
        requirements: tuple of Requirement
            Optional requirements explaining the rule's origin.

        Returns
        -------
        rule: PackageRule
        """
        return PackageRule((-self._pool.package_id(package),), reason,
                           requirements=requirements)

    # -------------------------------------------------
    # API to assemble individual rules from requirement
    # -------------------------------------------------
    def _add_rule(self, rule, rule_type):
        """
        Add the given rule to the internal rules set.

        Does nothing if the rule is None.

        Parameters
        ----------
        rule: PackageRule or None
            The rule to add
        rule_type: RuleType
            Rule's type
        """
        if rule is not None and rule not in self._rules_set:
            self._rules_set[rule] = None

    def _add_install_requires_rules(self, package, work_queue, requirements):
        all_dependency_candidates = []
        for constraints in package.install_requires:
            pkg_requirement = InstallRequirement.from_constraints(constraints)
            dependency_candidates = self._pool.what_provides(pkg_requirement)

            # We add our new requirement to the stack of requirements we've
            # gathered so far for these rules.
            combined_requirements = (
                requirements + (pkg_requirement,)
                if requirements is not None
                else None)

            if not dependency_candidates:
                pkg_msg = "'{0.name} {0.version}'"
                if hasattr(package, 'repository_info'):
                    pkg_msg += " from '{0.repository_info.name}'"
                pkg_str = pkg_msg.format(package)
                req_str = str(pkg_requirement)
                msg = ("Blocking package {0!s}: no candidates found for"
                       " dependency {1!r}").format(pkg_str, req_str)
                if self.strict:
                    # We only raise an exception if this comes directly from a
                    # job requirement. Unfortunately, we don't track that
                    # explicitly because we push all of the work through a
                    # queue. As a proxy, we can examine the associated
                    # requirements directly. Everything is rooted in a job, so
                    # if there's only one requirement, that must be it.
                    if len(requirements) == 1:
                        raise MissingInstallRequires(pkg_requirement, msg)
                    else:
                        logger.warning(msg)
                else:
                    logger.info(msg)

                rule = self._create_remove_rule(
                    package, RuleType.package_broken,
                    requirements=combined_requirements,
                )
                self._add_rule(rule, "package")
                return

            rule = self._create_dependency_rule(
                package, dependency_candidates, RuleType.package_requires,
                combined_requirements)
            self._add_rule(rule, "package")
            # We're "buffering" this so that we don't queue up any dependencies
            # unless they are all successfully processed
            all_dependency_candidates.extend(
                (candidate, combined_requirements)
                for candidate in dependency_candidates)
        work_queue.extend(all_dependency_candidates)

    def _add_conflicts_rules(self, package, requirements):
        """
        Create rules for each of the known conflicts with `package`.
        """

        # Conflicts due to same-name
        pkg_requirement = ConflictRequirement._from_string(package.name)
        obsolete_providers = self._pool.what_provides(pkg_requirement)
        # We add our new requirement to the stack of requirements we've
        # gathered so far for these rules.
        combined_requirements = (
            requirements + (pkg_requirement,)
            if requirements is not None
            else None)
        for provider in obsolete_providers:
            if provider != package and provider.name == package.name:
                reason = RuleType.package_same_name
                rule = self._create_conflicts_rule(
                    package, provider, reason, combined_requirements)
                self._add_rule(rule, "package")

        # Explicit conflicts in package metadata
        for constraints in package.conflicts:
            pkg_requirement = ConflictRequirement.from_constraints(constraints)
            conflict_providers = self._pool.what_provides(pkg_requirement)
            combined_requirements = (
                requirements + (pkg_requirement,)
                if requirements is not None
                else None)

            if not conflict_providers:
                pkg_msg = "'{0.name} {0.version}'"
                if hasattr(package, 'repository_info'):
                    pkg_msg += " from '{0.repository_info.name}'"
                pkg_str = pkg_msg.format(package)
                req_str = str(pkg_requirement)
                msg = ("No candidates found for requirement {0!r}, needed"
                       " for conflict with {1!s}").format(req_str, pkg_str)
                if self.strict:
                    # We only raise an exception if this comes directly from a
                    # job requirement. Unfortunately, we don't track that
                    # explicitly because we push all of the work through a
                    # queue. As a proxy, we can examine the associated
                    # requirements directly.
                    if len(requirements) == 1:
                        raise MissingConflicts(pkg_requirement, msg)
                    else:
                        logger.warning(msg)
                else:
                    # We just ignore missing constraints. They don't break
                    # anything.
                    logger.info(msg)

            for provider in conflict_providers:
                rule = self._create_conflicts_rule(
                    package, provider, RuleType.package_conflicts,
                    requirements=combined_requirements)
                self._add_rule(rule, "package")

    def _add_package_rules(self, package, requirements=None):
        """
        Create all the rules required to satisfy installing the given package.
        """
        work_queue = deque()
        work_queue.append((package, requirements))

        while len(work_queue) > 0:
            p, requirements = work_queue.popleft()
            p_id = self._pool.package_id(p)
            if p_id not in self.added_package_ids:
                self.added_package_ids.add(p_id)
                # We have to pass along our history-stack of requirements so
                # that they can be attached to the rules generated from here.
                self._add_install_requires_rules(p, work_queue, requirements)
                self._add_conflicts_rules(p, requirements)

    def _add_install_job_rules(self, job):
        packages = self._pool.what_provides(
            job.requirement, use_modifiers=False)
        if len(packages) > 0:
            for package in packages:
                # This is an optimization to avoid iterating over the installed
                # packages again.
                package_id = self._pool.package_id(package)
                if package_id not in self.installed_package_ids:
                    # Rules created directly from a job requirement have no
                    # other requirements in their history-stack
                    self._add_package_rules(
                        package, requirements=(job.requirement,))

            rule = self._create_install_one_of_rule(
                packages, RuleType.job_install,
                requirements=(job.requirement,))
            self._add_rule(rule, "job")
        else:
            raise NoPackageFound(job.requirement, str(job.requirement))

    def _add_remove_job_rules(self, job):
        packages = self._pool.what_provides(
            job.requirement, use_modifiers=False)
        for package in packages:
            rule = self._create_remove_rule(
                package, RuleType.job_remove, requirements=(job.requirement,))
            self._add_rule(rule, "job")

    def _add_update_job_rules(self, job):
        """
        Create rules that force the update of the package by requiring all of
        the standard rules then adding an additional rule for just the most
        recent version.
        """
        packages = self._pool.what_provides(
            job.requirement, use_modifiers=False)
        if len(packages) == 0:
            return

        # An update request *must* install the latest package version
        def key(package):
            package_id = self._pool.package_id(package)
            installed = package_id in self.installed_package_ids
            return (package.version, installed)
        package = max(packages, key=key)
        self._add_package_rules(package, requirements=(job.requirement,))
        rule = PackageRule(
            (self._pool.package_id(package),),
            RuleType.job_update,
            requirements=(job.requirement,),
        )
        self._add_rule(rule, "job")

    def _add_installed_package_rules(self, package):
        packages_all_versions = self._pool.name_to_packages(package.name)
        for other in packages_all_versions:
            self._add_package_rules(other)

    def _add_job_rules(self):
        for job in self.request.jobs:
            if job.kind in (JobType.install, JobType.soft_update):
                self._add_install_job_rules(job)
            elif job.kind == JobType.remove:
                self._add_remove_job_rules(job)
            elif job.kind == JobType.hard_update:
                self._add_update_job_rules(job)
            else:
                msg = "Job kind {0!r} not supported".format(job.kind)
                raise NotImplementedError(msg)



import unittest
from collections import OrderedDict

from okonomiyaki.versions import EnpkgVersion

from simplesat.errors import (
    InvalidConstraint, InvalidDependencyString, SolverException
)

from ..kinds import Equal
from ..multi import MultiConstraints
from ..requirement import InstallRequirement, parse_package_full_name
from ..constraint_modifiers import ConstraintModifiers, modify_requirement


R = InstallRequirement._from_string
V = EnpkgVersion.from_string


class TestRequirementFromConstraint(unittest.TestCase):

    def test_comparison(self):
        # Given
        constraints0 = ("numpy", ((">= 1.8.1-3", "< 1.9.0"),))
        constraints1 = ("numpy", ((">= 1.8.1-3", "< 1.9.1"),))

        # When
        requirement0 = InstallRequirement.from_constraints(constraints0)
        requirement1 = InstallRequirement.from_constraints(constraints1)

        # Then
        self.assertTrue(requirement0 != requirement1)

    def test_hashing(self):
        # Given
        constraints0 = ("numpy", ((">= 1.8.1-3", "< 1.9.1"),))

        # When
        requirement0 = InstallRequirement.from_constraints(constraints0)
        requirement1 = InstallRequirement.from_constraints(constraints0)

        # Then
        self.assertEqual(requirement0, requirement1)
        self.assertEqual(hash(requirement0), hash(requirement1))

    def test_any(self):
        # Given
        constraints0 = ("numpy", ((),))

        # When
        requirement = InstallRequirement.from_constraints(constraints0)

        # Then
        self.assertTrue(requirement.matches(V("1.8.1-2")))
        self.assertTrue(requirement.matches(V("1.8.1-3")))
        self.assertTrue(requirement.matches(V("1.8.2-1")))
        self.assertTrue(requirement.matches(V("1.9.0-1")))
        self.assertEqual(
            requirement,
            InstallRequirement.from_constraints(("numpy", (("*",),)))
        )

    def test_simple(self):
        # Given
        constraints0 = ("numpy", ((">= 1.8.1-3", "< 1.9.0"),))

        # When
        requirement = InstallRequirement.from_constraints(constraints0)

        # Then
        self.assertFalse(requirement.matches(V("1.8.1-2")))
        self.assertTrue(requirement.matches(V("1.8.1-3")))
        self.assertTrue(requirement.matches(V("1.8.2-1")))
        self.assertFalse(requirement.matches(V("1.9.0-1")))

    def test_multiple_fails(self):
        # Given
        constraints0 = (("numpy", ((">= 1.8.1-3",),)),
                        ("scipy", (("< 1.9.0",),)))

        # Then
        with self.assertRaises(InvalidConstraint):
            InstallRequirement.from_constraints(constraints0)

    def test_disjunction_fails(self):
        constraints0 = ("numpy", (("< 1.8.0",), (">= 1.8.1-3",)))

        # Then
        with self.assertRaises(InvalidConstraint):
            InstallRequirement.from_constraints(constraints0)

    def test_has_any_version_constraint(self):
        # Given
        requirements = [
            (("numpy", ((),)), False),
            (("numpy", (("*",),)), False),
            (("numpy", (("< 1.8.1",),)), True),
            (("numpy", (("== 1.8.1-1",),)), True),
            (("numpy", (("^= 1.8.1",),)), True),
        ]

        # When/Then
        for pretty_string, has_any_version_constraint in requirements:
            requirement = InstallRequirement.from_constraints(pretty_string)
            self.assertEqual(
                requirement.has_any_version_constraint,
                has_any_version_constraint
            )


class TestRequirementFromString(unittest.TestCase):
    def test_comparison(self):
        # Given
        requirement_string1 = "numpy >= 1.8.1-3, numpy < 1.9.0"
        requirement_string2 = "numpy >= 1.8.1-3, numpy < 1.9.1"

        # When
        requirement1 = R(requirement_string1)
        requirement2 = R(requirement_string2)

        # Then
        self.assertTrue(requirement1 != requirement2)

    def test_hashing(self):
        # Given
        requirement_string = "numpy >= 1.8.1-3, numpy < 1.9.0"

        # When
        requirement1 = R(requirement_string)
        requirement2 = R(requirement_string)

        # Then
        self.assertEqual(requirement1, requirement2)
        self.assertEqual(hash(requirement1), hash(requirement2))

    def test_any(self):
        # Given
        requirement_string = "numpy"
        r_requirement = InstallRequirement.from_constraints(
            ("numpy", (("*",),)))
        r_requirement_empty = InstallRequirement.from_constraints(
            ("numpy", ((),)))

        # When
        requirement = R(requirement_string)

        # Then
        self.assertTrue(requirement.matches(V("1.8.1-2")))
        self.assertTrue(requirement.matches(V("1.8.1-3")))
        self.assertTrue(requirement.matches(V("1.8.2-1")))
        self.assertTrue(requirement.matches(V("1.9.0-1")))
        self.assertEqual(requirement, r_requirement)
        self.assertEqual(requirement, r_requirement_empty)
        self.assertEqual(r_requirement, r_requirement_empty)

        # Given
        requirement_string = "numpy *"

        # When
        requirement = R(requirement_string)

        # Then
        self.assertEqual(requirement, r_requirement)
        self.assertEqual(requirement, r_requirement_empty)
        self.assertEqual(r_requirement, r_requirement_empty)

    def test_simple(self):
        # Given
        requirement_string = "numpy >= 1.8.1-3, numpy < 1.9.0"

        # When
        requirement = R(requirement_string)

        # Then
        self.assertFalse(requirement.matches(V("1.8.1-2")))
        self.assertTrue(requirement.matches(V("1.8.1-3")))
        self.assertTrue(requirement.matches(V("1.8.2-1")))
        self.assertFalse(requirement.matches(V("1.9.0-1")))

    def test_multiple_fails(self):
        # Given
        requirement_string = "numpy >= 1.8.1-3, scipy < 1.9.0"

        # When
        with self.assertRaises(InvalidDependencyString):
            R(requirement_string)

    def test_from_package_string(self):
        # Given
        package_s = "numpy-1.8.1-1"

        # When
        requirement = InstallRequirement.from_package_string(package_s)

        # Then
        self.assertEqual(requirement.name, "numpy")
        self.assertEqual(requirement._constraints,
                         MultiConstraints([Equal(V("1.8.1-1"))]))

    def test_has_any_version_constraint(self):
        # Given
        requirements = [
            ("numpy", False),
            ("numpy < 1.8.1", True),
            ("numpy == 1.8.1-1", True),
            ("numpy ^= 1.8.1", True),
        ]

        # When/Then
        for pretty_string, has_any_version_constraint in requirements:
            requirement = R(pretty_string)
            self.assertEqual(
                requirement.has_any_version_constraint,
                has_any_version_constraint
            )


class TestParsePackageFullName(unittest.TestCase):

    def test_simple(self):
        # Given
        for package_s in ("numpy-1.8.1-1", "numpy 1.8.1-1"):

            # When
            name, version = parse_package_full_name(package_s)

            # Then
            self.assertEqual(name, "numpy")
            self.assertEqual(version, "1.8.1-1")


class TestRequirement(unittest.TestCase):
    def test_repr(self):
        # Given
        constraints = (
            "numpy", (("^= 1.8.0",),)
        )
        r_repr = "InstallRequirement('numpy ^= 1.8.0')"

        # When
        requirement = InstallRequirement.from_constraints(constraints)

        # Then
        self.assertMultiLineEqual(repr(requirement), r_repr)

        # Given
        constraints = (
            "numpy", ((">= 1.8.0", "< 1.10.0"),)
        )
        r_repr = "InstallRequirement('numpy >= 1.8.0-0, < 1.10.0-0')"

        # When
        requirement = InstallRequirement.from_constraints(constraints)

        # Then
        self.assertMultiLineEqual(repr(requirement), r_repr)

        # Given
        constraints = (
            "numpy", ((">= 1.8.0-0", "< 1.10.0-0"),)
        )
        r_repr = "InstallRequirement('numpy >= 1.8.0-0, < 1.10.0-0')"

        # When
        requirement = InstallRequirement.from_constraints(constraints)

        # Then
        self.assertMultiLineEqual(repr(requirement), r_repr)


class TestRequirementModification(unittest.TestCase):

    CONSTRAINTS = (
        # Greater than
        "> 1.1.1-1", ">= 1.1.1-1",
        # Less than
        "< 1.1.1-1", "<= 1.1.1-1",
        # Equality
        "^= 1.1.1", "== 1.1.1-1", "!= 1.1.1-1", "*")

    TARGETS = {
        'allow_newer': (
            # Greater than
            "> 1.1.1-1", ">= 1.1.1-1",
            # Less than
            "*", "*",
            # Equality
            ">= 1.1.1", ">= 1.1.1-1", "!= 1.1.1-1", "*"),
        'allow_older': (
            # Greater than
            "*", "*",
            # Less than
            "< 1.1.1-1", "<= 1.1.1-1",
            # Equality
            "<= 1.1.1-999999999", "<= 1.1.1-1", "!= 1.1.1-1", "*"),
        'allow_any': ("*", "*", "*", "*", "*", "*", "!= 1.1.1-1", "*"),
    }

    def test_modify_single(self):
        for mode in ('allow_newer', 'allow_older', 'allow_any'):
            modifiers = self._make_modifiers(**{mode: 'A'})
            for before_c, after_c in zip(self.CONSTRAINTS, self.TARGETS[mode]):
                before = 'A ' + before_c
                after = 'A ' + after_c
                self.assertModification(before, after, modifiers)

    def test_modify_multi(self):
        # When the constraints are all together as a single requirement
        requirement_strings = ('A ' + c for c in self.CONSTRAINTS)
        before = ', '.join(requirement_strings)

        for mode in ('allow_newer', 'allow_older', 'allow_any'):
            target_requirement_strings = ('A ' + c for c in self.TARGETS[mode])
            after = ', '.join(self._stable_unique(target_requirement_strings))
            modifiers = self._make_modifiers(**{mode: 'A'})
            self.assertModification(before, after, modifiers)

    def test_newer_older_is_any(self):
        # When
        modifiers = self._make_modifiers(allow_newer='A', allow_older='A')
        targets_any = self.TARGETS['allow_any']
        for before_c, after_c in zip(self.CONSTRAINTS, targets_any):
            before = 'A ' + before_c
            after = 'A ' + after_c
            self.assertModification(before, after, modifiers)

    def test_collapse_multiple_any(self):
        # Given
        requirement = R("MKL >= 1.2.1-2, MKL != 2.3.1-1, MKL < 1.4")
        expected = R("MKL, MKL != 2.3.1-1")

        # When
        modifiers = ConstraintModifiers(allow_any='MKL')
        modified = modify_requirement(requirement, modifiers)
        constraints = modified._constraints._constraints

        # Then
        self.assertEqual(2, len(constraints))
        self.assertEqual(expected, modified)

    def _stable_unique(self, sequence):
        return tuple(OrderedDict.fromkeys(sequence).keys())

    def _make_modifiers(self, **kw):
        default = {
            'allow_any': ("B",),
            'allow_older': ("B",),
            'allow_any': ("B",),
        }
        default.update(**kw)
        return ConstraintModifiers(**default)

    def assertModification(self, before, after, modifiers):
        before_r = R(before)
        expected = R(after)
        result = modify_requirement(before_r, modifiers)
        msg = ("""
            before: {}
            expected: {}
            result: {}
            modifiers: {}
        """).format(before_r, expected, result, modifiers.asdict())
        self.assertEqual(expected, result, msg=msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import six


class AssignmentSet(object):

    """A collection of literals and their assignments."""

    def __init__(self, assignments=None):
        # Changelog is a dict of id -> (original value, new value)
        # FIXME: Verify that we really need ordering here
        self._data = {}
        self._orig = {}
        self._seen = set()
        self._cached_changelog = None
        self._assigned_ids = set()
        self.new_keys = set()
        for k, v in (assignments or {}).items():
            self[k] = v

    def __setitem__(self, key, value):

        abskey = abs(key)

        if abskey not in self._seen:
            self.new_keys.add(abskey)

        if value is None:
            del self[key]
        else:
            self._update_diff(key, value)
            self._data[key] = value
            self._data[-key] = not value
            self._assigned_ids.add(abs(key))

        self._seen.add(abskey)

    def __delitem__(self, key):
        self._seen.discard(abs(key))
        if key not in self._data:
            return
        self._update_diff(key, None)
        del self._data[key]
        del self._data[-key]
        self._assigned_ids.discard(abs(key))

    def __getitem__(self, key):
        val = self._data.get(key)
        if val is None and abs(key) not in self._seen:
            raise KeyError(key)
        return val

    def get(self, key, default=None):
        return self._data.get(key, default)

    def __len__(self):
        return len(self._seen)

    def __iter__(self):
        return iter(self.keys())

    def __contains__(self, key):
        return abs(key) in self._seen

    def items(self):
        return sorted(
            (k, self._data.get(k))
            for k in self._seen)

    def iteritems(self):
        return iter(self.items())

    def keys(self):
        return [k for k, _ in self.items()]

    def values(self):
        return [v for _, v in self.items()]

    def _update_diff(self, key, value):
        # This must be called before _data is updated
        if key < 0 and value is not None:
            key = -key
            value = not value
        prev = self._data.get(key)
        self._orig.setdefault(key, prev)
        # If a value changes, dump the cached changelog
        self._cached_changelog = None

    def get_changelog(self):
        if self._cached_changelog is None:
            self._cached_changelog = {
                key: (old, new)
                for key, old in six.iteritems(self._orig)
                for new in [self._data.get(key)]
                if new != old
            }
        return self._cached_changelog

    def consume_changelog(self):
        old = self.get_changelog()
        self._orig = {}
        self._cached_changelog = {}
        self.new_keys.clear()
        return old

    def copy(self):
        new = AssignmentSet()
        new._data = self._data.copy()
        new._orig = self._orig.copy()
        new._seen = self._seen.copy()
        new._assigned_ids = self._assigned_ids.copy()
        new.new_keys = self.new_keys.copy()
        return new

    def to_dict(self):
        return dict(self.items())

    def value(self, lit):
        """ Return the value of literal. """
        return self._data.get(lit)

    @property
    def num_assigned(self):
        return len(self._assigned_ids)

    @property
    def assigned_ids(self):
        return self._assigned_ids

    @property
    def unassigned_ids(self):
        return self._seen.difference(self._assigned_ids)



from __future__ import absolute_import

from .constraints import Requirement, InstallRequirement
from .package import PackageMetadata, RepositoryPackageMetadata, RepositoryInfo
from .pool import Pool
from .repository import Repository
from .request import JobType, Request

try:  # pragma: no cover
    from ._version import (
        version as __version__, version_info as __version_info__,
        is_released as __is_released__, git_revision as __git_revision__,
    )
except ImportError:  # pragma: no cover
    __is_released__ = False
    __version__ = __git_revision__ = "unknown"
    __version_info__ = (0, 0, 0, "unknown", 0)



from collections import defaultdict

import six

from simplesat import InstallRequirement, Pool
from simplesat.utils.graph import (package_lit_dependency_graph,
                                   transitive_neighbors)


def compute_dependencies(repositories, requirement, transitive=False):
    """ Compute packages in the given repos on which `requirement` depends.

    Parameters
    ----------------
    repositories : iterable of Repository objects
    requirement : Requirement
        The package requirement for which to compute dependencies.
    transitive : bool
        If True, recursively walk up the dependency graph. If False (default),
        only returns the packages on which the package directly depends.

    Returns
    -----------
    dependencies : set
        Set of packages in the given repositories that any packages satisfying
        the given requirement depend on.
    """
    pool = Pool(repositories)
    neighbors = _neighbors_in_repositories(pool, transitive)
    dependencies = _neighbors_for_requirement(pool, neighbors, requirement)
    return dependencies


def compute_reverse_dependencies(repositories, requirement, transitive=False):
    """ Compute packages in `repositories` that depend on the given requirement

    Parameters
    ----------------
    repositories : iterable of Repository objects
    requirement : Requirement
        The package requirement for which to compute reverse dependencies.
    transitive : bool
        If True, recursively walk up the dependency graph. If False (default),
        only returns the packages that directly depend on the package.

    Returns
    -----------
    dependencies : set
        Set of packages in the given repositories that depend on any of the
        packages satisfying the given requirement.
    """
    pool = Pool(repositories)
    reverse_neighbors = _reverse_neighbors_in_repositories(pool, transitive)
    dependencies = _neighbors_for_requirement(pool, reverse_neighbors,
                                              requirement)
    return dependencies


def compute_leaf_packages(repositories):
    """ Compute the leaf packages in `repositories`. Leaf packages are packages
    with no reverse dependencies.

    Parameters
    ----------------
    repositories : iterable of Repository objects

    Returns
    -----------
    dependencies : set
        Set of leaf packages in the given repositories.
    """
    pool = Pool(repositories)
    reverse_neighbors = _reverse_neighbors_in_repositories(pool)

    leaf_packages = set()
    for package in pool.iter_packages():
        package_string = package.name + "-" + str(package.version)
        requirement = InstallRequirement.from_package_string(package_string)
        dependencies = _neighbors_for_requirement(pool, reverse_neighbors,
                                                  requirement)
        if not dependencies:
            leaf_packages.add(package)

    return leaf_packages


def _neighbors_in_repositories(pool, transitive=False):
    """ Compute neighboring packages for all packages in a pool of
    repositories.

    Parameters
    ----------------
    pool: Pool
    transitive : bool
        If True, recursively walk up the dependency graph. If False (default),
        only returns the packages on which the package directly depends.

    Returns
    -----------
    neighbors : dict
         dict of all packages mapped to their neighbors.
    """
    package_ids = set(pool.iter_package_ids())
    neighbors = _compute_dependency_dict(pool, package_ids, transitive)
    return neighbors


def _reverse_neighbors_in_repositories(pool, transitive=False):
    """ Compute Reverse mapping of packages in a pool of repositories such that
    package ids point to the packages which depend on them.

    Parameters
    ----------------
    pool: Pool
    transitive : bool
        If True, recursively walk up the dependency graph. If False (default),
        only returns the packages on which the package directly depends.

    Returns
    -----------
    reverse_neighbors : dict
         dict of all packages mapped to packages that depend on them.
    """
    neighbors = _neighbors_in_repositories(pool, transitive)
    reverse_neighbors = _reverse_mapping(neighbors)
    return reverse_neighbors


def _compute_dependency_dict(pool, package_ids, transitive=False):
    """ Return mapping of package ids to their dependent package ids.

    If transitive is False, a package id will point to only its immediate
    dependencies. Otherwise, the mapping will include all recursive
    dependencies.
    """
    graph = package_lit_dependency_graph(pool, package_ids, closed=False)
    if transitive:
        return transitive_neighbors(graph)
    return graph


def _reverse_mapping(mapping):
    reversed_map = defaultdict(set)
    for key, vals in six.iteritems(mapping):
        for v in vals:
            reversed_map[v].add(key)

    return reversed_map


def _neighbors_for_requirement(pool, neighbor_mapping, requirement):
    """ Compute neighboring packages for all packages satisfying `requirement`

    Parameters
    ----------------
    pool: Pool
    neighbor_mapping : dict
        Mapping from package ids to "neighboring" package ids. These may be,
        for example, the packages which depend on a package or those on which
        the package depends.
    requirement : Requirement
        The package requirement for which to look up neighbors. All packages
        which satisfy the requirement will be used.

    Returns
    -----------
    neighbors : set
         Set of packages in the pool that are neighbors of packages which
         satisfy `requirement`.
    """
    neighbors = set()
    for package_id in _package_ids_satisfying_requirement(pool, requirement):
        neighbor_ids = neighbor_mapping[package_id]
        neighbors.update(pool.id_to_package(d_id) for d_id in neighbor_ids)

    return neighbors


def _package_ids_satisfying_requirement(pool, requirement):
    """ Yield package ids found in `pool` which satisfy `requirement`. """
    for package in pool.what_provides(requirement):
        yield pool.package_id(package)



from __future__ import absolute_import

import six

from .utils import DefaultOrderedDict
from simplesat.constraints import Requirement, modify_requirement
from simplesat.errors import InvalidConstraint


class Pool(object):
    """ A pool of repositories.

    The main feature of a pool is to search for every package matching a
    given requirement.

    Parameters
    ----------
    repositories : list of Repository, optional
        The repositories to query for packages.
    modifiers : ConstraintModifiers, optional
        If given, modify the requirements prior to querying.
    """

    def __init__(self, repositories=None, modifiers=None):
        self._id = 1
        self._repositories = []
        # FIXME Mar-9-2016: temporarily changing these names to catch places
        # that were using the private API. If it has been awhile and you feel
        # like everything is ok, you can remove these trailing underscores.
        self._package_to_id_ = {}
        self._id_to_package_ = {}
        self._packages_by_name_ = DefaultOrderedDict(list)

        self.modifiers = modifiers

        for repository in repositories or []:
            self.add_repository(repository)

    def add_repository(self, repository):
        """ Add the repository to this pool.

        Parameters
        ----------
        repository : Repository
            The repository to add
        """
        self._repositories.append(repository)
        for package in repository:
            current_id = self._id
            self._id += 1
            self._id_to_package_[current_id] = package
            self._package_to_id_[package] = current_id
            for constraints in package.provides:
                req = Requirement.from_constraints(constraints)
                if req.has_any_version_constraint:
                    msg = ('Version constraints are not supported for'
                           ' package.provides metadata: {}')
                    raise InvalidConstraint(msg.format(req))
                self._packages_by_name_[req.name].append(package)

    def what_provides(self, requirement, use_modifiers=True):
        """ Computes the list of packages fulfilling the given
        requirement.

        Parameters
        ----------
        requirement : Requirement
            The requirement to match candidates against.
        use_modifiers : bool
            If True, modify the requirement according to self.modifiers.

        Returns
        -------
        list of PackageMetadata
            The packages satisfying `requirement`.
        """
        ret = []
        if requirement.name in self._packages_by_name_:
            if use_modifiers:
                requirement = self.modify_requirement(requirement)
            for package in self._packages_by_name_[requirement.name]:
                if requirement.matches(package.version):
                    ret.append(package)
        return ret

    def modify_requirement(self, requirement):
        """Return requirement modified by the pool's ConstraintModifiers."""
        if self.modifiers:
            requirement = modify_requirement(requirement, self.modifiers)
        return requirement

    def package_id(self, package):
        """ Returns the 'package id' of the given package."""
        try:
            return self._package_to_id_[package]
        except KeyError:
            msg = "Package {0!r} not found in the pool.".format(package)
            raise ValueError(msg)

    def id_to_package(self, package_id):
        """ Returns the package of the given 'package id'."""
        try:
            return self._id_to_package_[package_id]
        except KeyError:
            msg = "Package ID {0!r} not found in the pool.".format(package_id)
            raise ValueError(msg)

    def id_to_string(self, package_id):
        """
        Convert a package id to a nice string representation.
        """
        package = self._id_to_package_[abs(package_id)]
        package_string = package.name + "-" + str(package.version)
        if package_id > 0:
            return "+" + package_string
        else:
            return "-" + package_string

    def name_to_packages(self, name):
        return tuple(self._packages_by_name_[name])

    def iter_packages(self):
        """ Iterate over all PackageMetadata objects. """
        for package in six.iterkeys(self._package_to_id_):
            yield package

    def iter_package_ids(self):
        """ Iterate over all package ids. """
        for pid in six.iterkeys(self._id_to_package_):
            yield pid

    @property
    def package_ids(self):
        return tuple(self._id_to_package_.keys())



import collections
import os

import six
import yaml

from okonomiyaki.versions import EnpkgVersion

from simplesat.constraints import PrettyPackageStringParser, Requirement
from simplesat.package import RepositoryInfo, RepositoryPackageMetadata
from simplesat.pool import Pool
from simplesat.repository import Repository
from simplesat.request import Request
from simplesat.rules_generator import RulesGenerator
from simplesat.transaction import (
    InstallOperation, RemoveOperation, UpdateOperation
)


HERE = os.path.dirname(__file__)


def generate_rules_for_requirement(pool, requirement, installed_map=None):
    """Generate CNF rules for a requirement.

    Parameters
    ----------
    pool : Pool
        A Pool of Repositories to use when fulfilling the requirement.
    requirement : Requirement
        The description of the package to be installed.

    Returns
    -------
    rules : list
        Package rules describing the given scenario.

    """
    request = Request()
    request.install(requirement)

    rules_generator = RulesGenerator(
        pool, request, installed_map=installed_map)
    rules = list(rules_generator.iter_rules())
    return rules


def packages_from_definition(packages_definition):
    parser = PrettyPackageStringParser(EnpkgVersion.from_string)

    return [
        parser.parse_to_package(line.strip())
        for line in packages_definition.splitlines() if line.strip()
    ]


def pool_and_repository_from_packages(packages):
    repository = Repository(packages_from_definition(packages))
    pool = Pool([repository])
    return pool, repository


def parse_package_list(packages):
    """ Yield PackageMetadata instances given an sequence  of pretty package
    strings.

    Parameters
    ----------
    packages : iterator
        An iterator of package strings (e.g.
        'numpy 1.8.1-1; depends (MKL ^= 10.3)').
    """
    parser = PrettyPackageStringParser(EnpkgVersion.from_string)

    for package_str in packages:
        package = parser.parse_to_package(package_str)
        full_name = "{0} {1}".format(package.name, str(package.version))
        yield full_name, package


def repository_factory(package_names, repository_info, reference_packages):
    repository = Repository()
    for package_name in package_names:
        package = reference_packages[package_name]
        package = RepositoryPackageMetadata(package, repository_info)
        repository.add_package(package)
    return repository


def remote_repository(yaml_data, packages):
    repository_info = RepositoryInfo(u"remote")
    package_names = yaml_data.get("remote", packages.keys())
    return repository_factory(package_names, repository_info, packages)


def installed_repository(yaml_data, packages):
    repository_info = RepositoryInfo(u"installed")
    package_names = yaml_data.get("installed", [])
    return repository_factory(package_names, repository_info, packages)


class Scenario(object):

    """
    A high level description of a scenario that should be solved.

    The Scenario class bundles together several important related
    pieces of data that together characterize a package management
    scenario. This includes a
    :class:`Request <simplesat.request.Request>`, a singular
    :class:`Repository <simplesat.repository.Repository>` representing
    packages that are currently installed
    and a list of :class:`Repository <simplesat.repository.Respository>`
    representing available packages.

    The key feature is the ability to create one from a human-readable
    yaml description::

        >>> Scenario.from_yaml(io.StringIO(u'''
        ...     packages:
        ...         - MKL 10.2-1
        ...         - MKL 10.3-1
        ...         - numpy 1.7.1-1; depends (MKL == 10.3-1)
        ...         - numpy 1.8.1-1; depends (MKL == 10.3-1)
        ...
        ...     request:
        ...         - operation: "install"
        ...           requirement: "numpy"
        ... ''')
    """

    @classmethod
    def from_yaml(cls, file_or_filename):
        if isinstance(file_or_filename, six.string_types):
            with open(file_or_filename) as fp:
                data = yaml.load(fp, Loader=_UnicodeLoader)
        else:
            data = yaml.load(file_or_filename, Loader=_UnicodeLoader)

        scenario_requests = data.get("request", [])

        marked = list(data.get("marked", []))

        request = Request()

        for kind, values in data.get("modifiers", {}).items():
            for value in values:
                getattr(request, kind)(value)

        update_all = False

        for s_request in scenario_requests:
            kind = s_request["operation"]
            if kind == 'update_all':
                update_all = True
                continue
            requirement = Requirement._from_string(s_request["requirement"])
            try:
                marked.remove(requirement.name)
            except ValueError:
                pass
            getattr(request, kind)(requirement)

        if update_all:
            request_job = request.hard_update
        else:
            request_job = request.install

        for package_str in marked:
            request_job(Requirement._from_string(package_str))

        decisions = data.get("decisions", {})

        operations = cls._operations_from_transaction_list(
            data.get("transaction", []))

        pretty_operations = cls._operations_from_transaction_list(
            data.get("pretty_transaction", []))

        failure = data.get('failure')

        packages = collections.OrderedDict(
            parse_package_list(data.get("packages", [])))

        return cls(packages, [remote_repository(data, packages)],
                   installed_repository(data, packages), request,
                   decisions, operations, pretty_operations, failure=failure)

    @staticmethod
    def _operations_from_transaction_list(transaction_ops):
        def P(p):
            return next(parse_package_list([p]))[1]

        operations = []
        for operation in transaction_ops:
            if operation["kind"] == "install":
                operations.append(InstallOperation(P(operation["package"])))
            elif operation["kind"] == "update":
                operations.append(UpdateOperation(P(operation["to"]),
                                                  P(operation["from"])))
            elif operation["kind"] == "remove":
                operations.append(RemoveOperation(P(operation["package"])))
            else:
                msg = "invalid operation kind {!r}".format(operation["kind"])
                raise ValueError(msg)
        return operations

    def __init__(self, packages, remote_repositories, installed_repository,
                 request, decisions, operations, pretty_operations,
                 failure=None):
        self.packages = packages
        self.remote_repositories = remote_repositories
        self.installed_repository = installed_repository
        self.request = request
        self.decisions = decisions
        self.operations = operations
        self.pretty_operations = pretty_operations
        self.failure = failure

    @property
    def failed(self):
        return self.failure is not None

    def print_solution(self, pool, positive_decisions):
        for package_id in sorted(positive_decisions):
            package = pool.id_to_package(package_id)
            print("{}: {} {}".format(package_id, package.name,
                                     package.full_version))


def construct_yaml_str(self, node):
    # Override the default string handling function to always
    # return unicode objects
    return self.construct_scalar(node)


class _UnicodeLoader(yaml.Loader):
    pass


_UnicodeLoader.add_constructor(u'tag:yaml.org,2002:str', construct_yaml_str)



from collections import OrderedDict

from attr import attr, attributes

from .constraints import InstallRequirement
from simplesat.utils.graph import toposort, package_lit_dependency_graph


@attributes
class Operation(object):
    package = attr()


@attributes
class UpdateOperation(Operation):
    source = attr()


@attributes
class InstallOperation(Operation):
    pass


@attributes
class RemoveOperation(Operation):
    pass


class Transaction(object):

    def __init__(self, pool, decisions, installed_package_ids):
        self._pool = pool
        self._local_packages = {
            self._package_key(package_id): package_id
            for package_id in installed_package_ids
        }
        self.operations = self._safe_operations(
            decisions, installed_package_ids)
        self.pretty_operations = self._as_pretty_operations(self.operations)

    def __iter__(self):
        return iter(self.operations)

    def __str__(self):
        return self.to_string(self.operations)

    def _package_key(self, package_id):
        package = self._pool.id_to_package(package_id)
        return (package.name, package.version)

    @staticmethod
    def to_string(operations):
        lines = []
        for operation in operations:
            if isinstance(operation, InstallOperation):
                lines.append("Installing:\n\t{}".format(operation.package))
            elif isinstance(operation, UpdateOperation):
                lines.append(
                    "Updating:\n\t{}\n\t{}".format(operation.source,
                                                   operation.package)
                )
            elif isinstance(operation, RemoveOperation):
                lines.append("Removing\n\t{}".format(operation.package))
            else:
                msg = "Unknown operation: {!r}".format(operation)
                raise ValueError(msg)

        return "\n".join(lines)

    def to_simple_string(self):
        S = "{0.name} ({0.version})".format
        lines = []
        for operation in self.operations:
            if isinstance(operation, InstallOperation):
                lines.append("Installing {}".format(S(operation.package)))
            elif isinstance(operation, UpdateOperation):
                lines.append(
                    "Updating {} to {}".format(S(operation.source),
                                               S(operation.package))
                )
            elif isinstance(operation, RemoveOperation):
                lines.append("Removing {}".format(S(operation.package)))
            else:
                msg = "Unknown operation: {!r}".format(operation)
                raise ValueError(msg)

        return "\n".join(lines)

    def _as_pretty_operations(self, operations):
        pkg_to_ops = OrderedDict((op.package, [op]) for op in operations)

        for pkg in reversed(tuple(pkg_to_ops.keys())):
            if pkg in pkg_to_ops:
                for update in self._find_other_providers(pkg):
                    pkg_to_ops[pkg] += pkg_to_ops.pop(update, [])

        combine = self._merge_operations
        return [combine(ops) for ops in pkg_to_ops.values()]

    def _merge_operations(self, ops):
        if len(ops) == 1:
            return ops[0]
        rank = (InstallOperation, RemoveOperation)
        first, second = sorted(ops, key=lambda o: rank.index(o.__class__))
        return UpdateOperation(first.package, second.package)

    def _safe_operations(self, decisions, installed_package_ids):
        graph = package_lit_dependency_graph(self._pool, decisions,
                                             closed=True)
        removals = []
        installs = []

        # This builds from the bottom (no dependencies) up
        for group in toposort(graph):
            # Sort the set of independent packages for determinism
            for package_id in sorted(group, key=abs):
                assert package_id in decisions
                if package_id < 0 and -package_id in installed_package_ids:
                    removals.append(-package_id)
                elif (package_id > 0 and
                      package_id not in installed_package_ids):
                    installs.append(package_id)

        install_operations = []
        remove_operations = []

        # Installations should happen bottom up
        for package_id in installs:
            # The solver may suggest removing a currently installed package and
            # installing another version of that same package from a different
            # repo. Filter out operations which would install the same version
            # of a currently installed package and discard the associated
            # remove operation.
            key = self._package_key(package_id)
            preferred_id = self._local_packages.get(key, package_id)

            if preferred_id != package_id:
                # Because we are not installing the remote package, do not
                # remove original version
                if preferred_id in removals:
                    removals.remove(preferred_id)
            else:
                package = self._pool.id_to_package(package_id)
                install_operations.append(InstallOperation(package))

        # Removals should happen top down
        for package_id in reversed(removals):
            package = self._pool.id_to_package(package_id)
            remove_operations.append(RemoveOperation(package))

        return remove_operations + install_operations

    def _find_other_providers(self, package):
        # NOTE: this assumes that the name of the package is also the name of
        # the thing that is being provided. This is not always true. Consider
        # that apache2 and nginx can both provide "webserver", etc.
        requirement = InstallRequirement._from_string(package.name)
        return [
            p for p in self._pool.what_provides(requirement) if p != package
        ]



import re
from collections import defaultdict

from simplesat.package import PackageMetadata, ConstraintKinds
from .parser import _DISTRIBUTION_R, _VERSION_R, _WS_R


_WS_RS = _WS_R
_MAYBE_WS_RS = " *"
_DISTRIBUTION_RS = "(?P<distribution>{})".format(_DISTRIBUTION_R)
_VERSION_RS = "(?P<version>{})".format(_VERSION_R)
_CONSTRAINT_RS = "(?P<constraint>[^,]*)"

CONSTRAINT_BLOCK_RC = re.compile("(?P<kind>\w+)\s*\((?P<constraints>.*?)\)")
PACKAGE_RC = re.compile(_DISTRIBUTION_RS + _WS_RS + _VERSION_RS)
CONSTRAINT_RC = re.compile(_DISTRIBUTION_RS + _MAYBE_WS_RS + _CONSTRAINT_RS)

CONSTRAINT_SYNONYMS = {
    'depends': ConstraintKinds.install_requires,
    'install_requires': ConstraintKinds.install_requires,
    'conflicts': ConstraintKinds.conflicts,
    'provides': ConstraintKinds.provides,
}


class PrettyPackageStringParser(object):

    def __init__(self, version_factory):
        self._version_factory = version_factory

    def parse(self, pretty_string):
        """Return the dict representation of the pretty package string.

        Pretty package strings are of the form::

            numpy 1.8.1-1; install_requires (MKL == 10.3, nose ^= 1.3.4); conflicts (numeric); provides (numeric)  # noqa
        """
        pretty_string = pretty_string.strip()
        pkg = {}

        try:
            preamble, constraints_blocks = pretty_string.split(";", 1)
        except ValueError:
            preamble = pretty_string
            constraints_blocks = ''

        for match in CONSTRAINT_BLOCK_RC.finditer(constraints_blocks):
            kind = match.group('kind')
            constraints_str = match.group('constraints')
            if kind not in CONSTRAINT_SYNONYMS:
                msg = "Invalid package string. Unknown constraint kind: {!r}"
                raise ValueError(msg.format(kind))
            kind = CONSTRAINT_SYNONYMS[kind].value
            constraints = defaultdict(lambda: [[]])
            for match in CONSTRAINT_RC.finditer(constraints_str):
                dist = match.group('distribution')
                constraint_str = match.group('constraint')
                constraints[dist][0].append(constraint_str)
            pkg[kind] = constraints

        # Turn constraints into immutable nested tuples
        pkg = {
            kind: tuple(sorted(
                (dist, tuple(tuple(clist) for clist in constraints))
                for dist, constraints in dist_constraints.items()
            ))
            for kind, dist_constraints in pkg.items()
        }

        distribution, version = _parse_preamble(preamble)
        pkg["distribution"] = distribution
        pkg["version"] = self._version_factory(version)

        return pkg

    def parse_to_package(self, package_string):
        """ Parse the given pretty package string.

        Parameters
        ----------
        package_string : str
            The pretty package string, e.g.
            "numpy 1.8.1-1; depends (MKL == 10.3, nose ^= 1.3.4)"

        Returns
        -------
        package : PackageMetadata
        """
        pkg_dict = self.parse(package_string)
        distribution = pkg_dict.pop('distribution')
        version = pkg_dict.pop('version')
        return PackageMetadata(distribution, version, **pkg_dict)


def constraints_to_pretty_strings(constraint_tuples):
    """ Convert a sequence of constraint tuples as used in PackageMetadata to a
    list of pretty constraint strings.

    Parameters
    ----------
    constraint_tuples : tuple of constraint
        Sequence of constraint tuples, e.g. (("MKL", (("< 11", ">= 10.1"),)),)
    """
    flat_strings = [
        "{} {}".format(dist, constraint_string).strip()
        for dist, disjunction in constraint_tuples
        for conjunction in disjunction
        for constraint_string in conjunction
    ]
    return flat_strings


def package_to_pretty_string(package):
    """ Given a PackageMetadata instance, returns a pretty string."""
    template = "{0.name} {0.version}"
    constraint_kinds = (
        (ConstraintKinds.install_requires, package.install_requires),
        (ConstraintKinds.conflicts, package.conflicts),
        (ConstraintKinds.provides, package.provides),
    )
    for constraint_kind, constraints in constraint_kinds:
        # FIXME: perhaps 'provides' just shouldn't include the package name
        if constraint_kind == ConstraintKinds.provides:
            constraints = tuple((dist, disjunction)
                                for dist, disjunction in constraints
                                if dist != package.name)
        if len(constraints) > 0:
            string = ', '.join(constraints_to_pretty_strings(constraints))
            template += "; {} ({})".format(constraint_kind.value, string)
    return template.format(package)


def _parse_preamble(preamble):
    msg = "Invalid preamble: {0!r}".format(preamble)
    match = PACKAGE_RC.match(preamble)
    if not match:
        raise ValueError(msg)
    if match.span()[-1] != len(preamble):
        msg = msg + ", {!r}".format(match.group(0))
        raise ValueError(msg)
    groups = match.groupdict()
    return groups['distribution'], groups['version']



import enum

from attr import attr, attributes, Factory
from attr.validators import instance_of, optional

from .constraints import Requirement, ConstraintModifiers


@enum.unique
class JobType(enum.Enum):
    install = 1
    remove = 2
    soft_update = 3
    hard_update = 4
    upgrade = 5


@attributes
class _Job(object):
    requirement = attr(validator=optional(instance_of(Requirement)))
    kind = attr(validator=instance_of(JobType))

    def __attrs_post_init__(self):
        if self.requirement is None and self.kind != JobType.upgrade:
            raise ValueError(
                u"_Job requirement cannot be none if kind != {}",format(
                    JobType.upgrade
                )
            )

    def __str__(self):
        return u"{} {}".format(self.kind.name, self.requirement)


@attributes
class Request(object):
    """
    A proposed change to the state of the installed repository.

    The Request is built up from :class:`Requirement` objects and
    ad-hoc package constraint modifiers.

    Parameters
    ----------
    modifiers : ConstraintModifiers, optional
        The contraint modifiers are used to relax constraints when deciding
        on which packages meet a requirement.


    >>> from simplesat.request import Request
    >>> from simplesat.constraints import Requirement
    >>> request = Request()
    >>> recent_mkl = Requirement.from_string('MKL >= 11.0')
    >>> request.install(recent_mkl)
    >>> request.jobs
    [_Job(requirement=Requirement('MKL >= 11.0-0'), kind=<JobType.install: 1>)]
    >>> request.modifiers
    ConstraintModifiers(allow_newer=set(), allow_any=set(), allow_older=set())
    >>> request.allow_newer('MKL')
    >>> request.modifiers.asdict()
    {'allow_older': [], 'allow_any': ['MKL'], 'allow_newer': []}
    """

    modifiers = attr(default=Factory(ConstraintModifiers),
                     validator=instance_of(ConstraintModifiers))
    jobs = attr(default=Factory(list))

    def install(self, requirement):
        self._add_job(requirement, JobType.install)

    def remove(self, requirement):
        self._add_job(requirement, JobType.remove)

    def upgrade(self):
        self._add_job(None, JobType.upgrade)

    def hard_update(self, requirement):
        self._add_job(requirement, JobType.hard_update)

    def soft_update(self, requirement):
        self._add_job(requirement, JobType.soft_update)

    def allow_newer(self, package_name):
        self.modifiers.allow_newer.add(package_name)

    def allow_any(self, package_name):
        self.modifiers.allow_any.add(package_name)

    def allow_older(self, package_name):
        self.modifiers.allow_older.add(package_name)

    def _add_job(self, requirement, job_type):
        if len(self.jobs) > 0:
            if (job_type is JobType.upgrade or
                any(job.kind == JobType.upgrade for job in self.jobs)
            ):
                raise ValueError(
                    u"Requests with upgrade job can only have one job")
        self.jobs.append(_Job(requirement, job_type))



from .package_parser import PrettyPackageStringParser
from .requirement import (
    Requirement, ConflictRequirement, InstallRequirement
)
from .constraint_modifiers import (
    ConstraintModifiers, modify_requirement,
)



from __future__ import absolute_import

import bisect
import operator
import six

from .errors import NoPackageFound
from simplesat.constraints.requirement import Requirement


class Repository(object):
    """
    A Repository is a set of packages that knows about which package it
    contains.

    It also supports the iterator protocol. Iteration is guaranteed to be
    deterministic and independent of the order in which packages have been
    added.

    Parameters
    ----------
    packages : list of PackageMetadata
        The packages available in this repository.


    >>> from simplesat.constraints.package_parser import \\
    ...     pretty_string_to_package as P
    >>> mkl = P('MKL 10.3-1')
    >>> numpy1921 = P('numpy 1.9.2-1; depends (MKL)')
    >>> numpy1922 = P('numpy 1.9.2-2; depends (MKL, libgfortran)')
    >>> repository = Repository([mkl, numpy1922])
    >>> repository.add_package(numpy1921)
    >>> assert list(repository) == some_pkgs + [another_one]
    >>> numpies = repository.find_packages['numpy']
    >>> assert numpies == [numpy1921, numpy1922]
    """
    def __init__(self, packages=None):
        self._name_to_packages = {}
        self._default_factory = lambda: []
        # Sorted list of keys in self._name_to_packages, to keep iteration
        # over a repository reproducible
        self._names = []

        packages = packages or []
        for package in packages:
            self.add_package(package)

    def __len__(self):
        return sum(
            len(packages)
            for packages in six.itervalues(self._name_to_packages)
        )

    def __contains__(self, package_metadata):
        if package_metadata.name in self._name_to_packages:
            return (
                package_metadata
                in self._name_to_packages[package_metadata.name]
            )
        else:
            return False

    def __iter__(self):
        for name in self._names:
            for package in self._name_to_packages[name]:
                yield package

    def add_package(self, package_metadata):
        """ Add the given package to this repository.

        Parameters
        ----------
        package : PackageMetadata
            The package metadata to add. May be a subclass of PackageMetadata.

        Note
        ----
        If the same package is added multiple times to a repository, every copy
        will be available when calling find_package or when iterating.
        """
        if package_metadata.name not in self._name_to_packages:
            bisect.insort(self._names, package_metadata.name)

        self._name_to_packages.setdefault(
            package_metadata.name, self._default_factory()
        )
        self._name_to_packages[package_metadata.name].append(package_metadata)
        # Fixme: this should not be that costly as long as we don't have
        # many versions for a given package.
        self._name_to_packages[package_metadata.name].sort(
            key=operator.attrgetter("version")
        )

    def find_package(self, name, version):
        """Search for the first match of a package with the given name and
        version.

        Parameters
        ----------
        name : str
            The package name to look for.
        version : EnpkgVersion
            The version to look for.

        Returns
        -------
        package : PackageMetadata
            The corresponding metadata.
        """
        candidates = self._name_to_packages.get(name, self._default_factory())
        for candidate in candidates:
            if candidate.version == version:
                return candidate
        package_string = '{0}-{1}'.format(name, str(version))
        raise NoPackageFound(
            Requirement.from_package_string(package_string),
            "Package '{0}' not found".format(package_string),
        )

    def find_packages(self, name):
        """ Returns an iterable of package metadata with the given name, sorted
        from lowest to highest version.

        Parameters
        ----------
        name : str
            The package's name

        Returns
        -------
        packages : iterable
            Iterable of PackageMetadata instances (order is from lower to
            higher version)
        """
        return tuple(self._name_to_packages.get(name, self._default_factory()))

    def update(self, iterable):
        """ Add the packages from the given iterable into this repository.

        Parameters
        ----------
        """
        for package in iterable:
            self.add_package(package)



import sys
import unittest

from okonomiyaki.versions import EnpkgVersion

from simplesat.constraints.package_parser import (
    PrettyPackageStringParser, package_to_pretty_string
)
from simplesat.constraints.requirement import Requirement
from simplesat.package import PackageMetadata
from simplesat.errors import InvalidConstraint


V = EnpkgVersion.from_string


class TestPrettyPackageStringParser(unittest.TestCase):
    def test_invalid_formats(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = ""
        r_message = "Invalid preamble: "

        # Then
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

        # Given
        package_string = "numpy"
        r_message = "Invalid preamble: 'numpy'"

        # Then
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

        # Given
        package_string = "numpy1.8.0-1"
        r_message = ("Invalid preamble: ")

        # Then
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

        # Given
        package_string = "numpy 1.8.0-1 depends (nose >= 1.3.2)"
        r_message = ("Invalid preamble: ")

        # Then
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

        # Given
        package_string = "numpy; depends (nose >= 1.3.2)"
        r_message = ("Invalid preamble: ")

        # Then
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

        # Given
        package_string = "numpy 1.8.0-1; disparages (nose >= 1.3.2)"
        r_message = ("Invalid package string. "
                     "Unknown constraint kind: 'disparages'")

        # When
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

    def test_depends_simple(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = "numpy 1.8.0-1; depends (nose == 1.3.4-1)"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        install_requires = dict(package['install_requires'])

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, V("1.8.0-1"))
        self.assertTrue("nose" in install_requires)
        self.assertEqual(install_requires["nose"], (('== 1.3.4-1',),))

    def test_conflicts_simple(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = "numpy 1.8.0-1; conflicts (browsey ^= 1.3.0)"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        conflicts = dict(package['conflicts'])

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, V("1.8.0-1"))
        self.assertTrue("browsey" in conflicts)
        self.assertEqual(conflicts["browsey"], (('^= 1.3.0',),))

    def test_unversioned(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = "numpy 1.8.0-1; depends (nose, matplotlib == 1.3.2-1)"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        install_requires = dict(package['install_requires'])

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, V("1.8.0-1"))
        self.assertTrue("nose" in install_requires)
        self.assertEqual(install_requires["nose"], (('',),))
        self.assertEqual(install_requires["matplotlib"], (('== 1.3.2-1',),))

        # Given
        package_string = "numpy 1.8.0-1; depends (nose *, zope == 1.3.2-1)"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        install_requires = dict(package['install_requires'])

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, V("1.8.0-1"))
        self.assertTrue("nose" in install_requires)
        self.assertEqual(install_requires["nose"], (('*',),))
        self.assertEqual(install_requires["zope"], (('== 1.3.2-1',),))

    def test_special_characters(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = "shiboken_debug 1.2.2-5"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']

        # Then
        self.assertEqual(name, "shiboken_debug")
        self.assertEqual(version, V("1.2.2-5"))

        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = '; '.join((
            "scikits.image 0.10.0-1",
            "depends (scipy ^= 0.14.0, pil, zope.distribution *)"
        ))
        r_install_requires = (
            ('pil', (('',),)),
            ('scipy', (('^= 0.14.0',),)),
            ('zope.distribution', (('*',),))
        )

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        install_requires = package['install_requires']

        # Then
        self.assertEqual(name, "scikits.image")
        self.assertEqual(version, V("0.10.0-1"))
        self.assertEqual(install_requires, r_install_requires)

    def test_multiple(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = "numpy 1.8.0-1; depends (nose => 1.3, nose < 1.4)"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        install_requires = dict(package['install_requires'])

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, V("1.8.0-1"))
        self.assertIn("nose", install_requires)
        self.assertEqual(install_requires["nose"], (('=> 1.3', '< 1.4'),))

    def test_no_constraints(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = "numpy 1.8.0-1"

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, V("1.8.0-1"))
        self.assertEqual(len(package), 2)

    def test_complicated(self):
        # Given
        parse = PrettyPackageStringParser(V).parse
        package_string = '; '.join((
            "bokeh 0.2.0-3",
            "depends (numpy ^= 1.8.0, MKL == 10.3, requests >= 0.2)",
            # Intentional typo
            "conflits (bokeh-git, requests ^= 0.2.5, requests > 0.4)",
        ))
        r_message = ("Invalid package string. "
                     "Unknown constraint kind: 'conflits'")

        # Then
        with self.assertRaisesRegexp(ValueError, r_message):
            parse(package_string)

        # Given
        package_string = '; '.join((
            "bokeh_git 0.2.0-3",
            "install_requires (zope *, numpy ^= 1.8.0, requests >= 0.2)",
            "conflicts (requests ^= 0.2.5, requests > 0.4, bokeh)",
            "provides (webplot ^= 0.1, bokeh)",
        ))
        r_install_requires = (
            ("numpy", (("^= 1.8.0",),)),
            ("requests", ((">= 0.2",),)),
            ("zope", (("*",),)))
        r_conflicts = (
            ("bokeh", (('',),)),
            ("requests", (("^= 0.2.5", "> 0.4"),)))
        r_provides = (
            ("bokeh", (('',),)),
            ("webplot", (("^= 0.1",),)))

        # When
        package = parse(package_string)
        name = package['distribution']
        version = package['version']
        install_requires = package['install_requires']
        conflicts = package['conflicts']
        provides = package['provides']

        # Then
        self.assertEqual(name, "bokeh_git")
        self.assertEqual(version, V("0.2.0-3"))
        self.assertEqual(install_requires, r_install_requires)
        self.assertEqual(conflicts, r_conflicts)
        self.assertEqual(provides, r_provides)


class TestPackagePrettyString(unittest.TestCase):

    def test_simple(self):
        # Given
        package = PackageMetadata(u"numpy", V("1.8.1-1"))
        r_pretty_string = u"numpy 1.8.1-1"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

    def test_install_requires(self):
        # Given
        install_requires = (("MKL", (("== 10.3-1",),)),)
        package = PackageMetadata(u"numpy", V("1.8.1-1"), install_requires)

        r_pretty_string = u"numpy 1.8.1-1; install_requires (MKL == 10.3-1)"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

        # Given
        install_requires = (("nose", (("",),)),)
        package = PackageMetadata(u"numpy", V("1.8.1-1"), install_requires)

        r_pretty_string = "numpy 1.8.1-1; install_requires (nose)"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

    def test_conflicts(self):
        # Given
        conflicts = (("MKL", (("== 10.3-1",),)),)
        package = PackageMetadata(u"numpy", V("1.8.1-1"), conflicts=conflicts)

        r_pretty_string = u"numpy 1.8.1-1; conflicts (MKL == 10.3-1)"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

        # Given
        conflicts = (("nose", (("",),)),)
        package = PackageMetadata(u"numpy", V("1.8.1-1"), conflicts=conflicts)

        r_pretty_string = "numpy 1.8.1-1; conflicts (nose)"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

    def test_provides(self):
        # Given
        provides = ((u"dance", ((u">= 10.3-1",),)),)
        package = PackageMetadata(u"zumba", V("1.8.1-1"), provides=provides)

        r_pretty_string = u"zumba 1.8.1-1; provides (dance >= 10.3-1)"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

        # Given
        provides = (("cardio", (("*",),)),)
        package = PackageMetadata(u"zumba", V("1.8.1-1"), provides=provides)

        r_pretty_string = "zumba 1.8.1-1; provides (cardio *)"

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)

    def test_complicated(self):
        # Given
        install_requires = (
            ("numpy", (("^= 1.8.0",),)),
            ("requests", ((">= 0.2",),)),
            ("zope", (("*",),)))
        conflicts = (
            ("bokeh", (('',),)),
            ("requests", ((">= 0.2.5", "< 0.4"),)))
        provides = (
            ("bokeh", (('*',),)),)
        package = PackageMetadata(
            u"bokeh_git", V("0.2.0-3"),
            install_requires=install_requires,
            conflicts=conflicts,
            provides=provides)

        r_pretty_string = '; '.join((
            "bokeh_git 0.2.0-3",
            "install_requires (numpy ^= 1.8.0, requests >= 0.2, zope *)",
            "conflicts (bokeh, requests >= 0.2.5, requests < 0.4)",
            "provides (bokeh *)",
        ))

        # When
        pretty_string = package_to_pretty_string(package)

        # Then
        self.assertEqual(pretty_string, r_pretty_string)


class TestToPackage(unittest.TestCase):

    def test_simple(self):
        # Given
        s = u"zope.deprecated_ 2"
        parser = PrettyPackageStringParser(V)

        # When
        package = parser.parse_to_package(s)

        # Then
        self.assertEqual(package.name, "zope.deprecated_")
        self.assertEqual(package.version, V('2'))
        self.assertEqual(package.install_requires, ())

    def test_with_depends(self):
        # Given
        s = u"numpy 1.8.1; depends (MKL ^= 10.3)"
        parser = PrettyPackageStringParser(V)

        # When
        package = parser.parse_to_package(s)

        # Then
        self.assertEqual(package.name, "numpy")
        self.assertEqual(package.version, V('1.8.1'))
        self.assertEqual(package.install_requires, (("MKL", (("^= 10.3",),)),))

    def test_with_conflicts(self):
        # Given
        s = u"numpy 1.8.1; conflicts (MKL <= 10.3)"
        parser = PrettyPackageStringParser(V)

        # When
        package = parser.parse_to_package(s)

        # Then
        self.assertEqual(package.name, "numpy")
        self.assertEqual(package.version, V('1.8.1'))
        self.assertEqual(package.conflicts, (("MKL", (("<= 10.3",),)),))

    def test_with_provides(self):
        # Given
        s = u"numpy 1.8.1-4; provides (numeric)"
        parser = PrettyPackageStringParser(V)

        # When
        package = parser.parse_to_package(s)

        # Then
        self.assertEqual(package.name, "numpy")
        self.assertEqual(package.version, V('1.8.1-4'))
        self.assertEqual(package.provides,
                         (('numpy', (('*',),)),
                          ("numeric", (("",),))))

    def test_complicated(self):
        # Given
        install_requires = (
            ("numpy", (("^= 1.8.0",),)),
            ("requests", ((">= 0.2",),)),
            ("zope", (("*",),)))
        conflicts = (
            ("bokeh", (('',),)),
            ("requests", ((">= 0.2.5", "< 0.4"),)))
        provides = (
            ("bokeh", (('*',),)),)
        expected = PackageMetadata(
            u"bokeh_git", V("0.2.0-3"),
            install_requires=install_requires,
            conflicts=conflicts,
            provides=provides)
        parser = PrettyPackageStringParser(V)

        # When
        s = '; '.join((
            "bokeh_git 0.2.0-3",
            "install_requires (numpy ^= 1.8.0, requests >= 0.2, zope *)",
            "conflicts (bokeh, requests >= 0.2.5, requests < 0.4)",
            "provides (bokeh *)",
        ))
        result = parser.parse_to_package(s)

        # Then
        self.assertEqual(result, expected)


class TestParseScaryPackages(unittest.TestCase):

    # NOTE: gathered with 'scripts/index_to_scary_package_strings.py'
    SCARY_PACKAGE_STRINGS = """
        7z 9.20-1
        wb 19-4
        qt 2.5-1
        pytz 19-4
        ply 214-1
        _pylor 8a-0
        pylor_ 4-2
        sympy 210-1
        pycdf 210-3
        zdaemon 210-4
        argparse 210-5
        setupdocs 214-2
        z3c.amf 3.0a1-1
        traitsgui 210-2
        ua2.djfab 0.1a1-1
        z3ext.seo 2.0b8-1
        tw2.core 1.2.0a1-1
        lck.i18n 1.2.0a1-1
        tw2.jquery 1.2.0a1-1
        p01.fswidget 1.0b2-1
        multiprocessing 214-2
        z3ext.content 2.0b1-1
        webapp2_static 1.0b2-1
        z3c.offlinepack 1.0b4-1
        js.jquery_jqote2 1.0b3-1
        p01.recipe.setup 1.0a3-1
        js.jquery_jqote2 0.1a1-1
        z3c.checkversions 0.1a2-1
        z3ext.preferences 1.0b7-1
        js.jquery_jqote2 2.0b11-1
        z3c.repoexternals 1.0b1-1
        js.jquery_jqote2 1.2.0a1-1
        py_1digit_checksum 1.0b2-1
        sphinxjp.themes.s6 4.5b1-1
        z3c.layer.ready2go 0.1b8-1
        z3ext.content.forms 1.2.0a1-1
        z3ext.portlets.recent 0.1.0a2-1
        o2w_cache_invalidator 1.2.0a1-1
        z3c.setuptools_mercurial 1.0b1-1
        isotoma.depends.plone4_1 1.0b4-1
        z3c.setuptools_mercurial 1.7b5-1
        isotoma.depends.plone4_1 1.0a1-1
        z3c.setuptools_mercurial 1.0a2-1
        example.rtsubsites_theme 1.0rc4-1
        isotoma.depends.zope2_13_8 2.0b7-1
        isotoma.depends.plone4_1 1.5.0b1-1
        isotoma.depends.zope2_13_8 1.2b1-1
        isotoma.depends.zope2_13_8 1.7b4-1
        z3c.setuptools_mercurial 2.0.1b1-1
        c2.search.customdescription 2.0b2-1
        isotoma.depends.zope2_13_8 1.3rc5-1
        tiddlywebplugins.sqlalchemy2 1.0b2-1
        products.zope_hotfix_20110622 0.1a1-1
        plonetheme.greenearththeme3_0 2.0b9-1
        plonetheme.greenearththeme3_0 0.1a1-1
        products.zope_hotfix_20110622 3.0a1-1
        plonetheme.greenearththeme3_0 1.0b3-1
        products.zope_hotfix_20111024 0.2a1-1
        isotoma.depends.plone4_1 4.3.0.dev80-1
        isotoma.recipe.zope2instance 1.2.0a1-1
        products.zope_hotfix_20111024 0.1b12-1
        products.zope_hotfix_20111024 2.1.1a1-1
        products.zope_hotfix_20111024 1.2.0a1-1
        d51.django.virtualenv.test_runner 1.0b1-1
        products.zope_hotfix_cve_2010_3198 0.1a1-1
        products.zope_hotfix_cve_2010_1104 1.0a5-1
        products.zope_hotfix_cve_2011_3587 2.0b4-1
        d51.django.virtualenv.test_runner 1.0rc11-1
        d51.django.virtualenv.test_runner 1.2.0a1-1
        products.zope_hotfix_cve_2011_3587 0.8.0b1-1
        collective.z3cform.datagridfield_demo 2.0b9-1
        collective.z3cform.datagridfield_demo 1.2.0a1-1
        collective.z3cform.datagridfield_demo 2.5.4rc-1
        collective.z3cform.datagridfield_demo 1.3a7.1-1
        products.zope_hotfix_cve_2010_3198 4.3.0.dev80-1
        collective.z3cform.datagridfield_demo 4.3.0.dev29-1
        collective.blueprint.translationlinker 0.10.2.dev4396-1
        trytond_account_invoice_line_standalone 0.10.2.dev4365-1
        pyf.components.postprocess.email_sender 0.12.0.dev4791-1
        pyf.components.consumers.xhtmlpdfwriter 0.12.0.dev4716-1
        trytond_purchase_invoice_line_standalone 0.12.0.dev4681-1
        experimental.daterangeindexoptimisations 0.11.0.dev4622-1
        trytond_purchase_invoice_line_standalone 0.12.4.dev4948-1
        harobed.paster_template.advanced_package 0.13.0.dev5015-1
        trytond_purchase_invoice_line_standalone 0.11.0.dev4638-1
        trytond_purchase_invoice_line_standalone 2.0.4.dev25071-1
        trytond_purchase_invoice_line_standalone 0.11.0.dev4629-1
        harobed.paster_template.advanced_package 2.0.4.dev25079-1
        harobed.paster_template.advanced_package 0.12.0.dev4690-1
        trytond_purchase_invoice_line_standalone 0.11.0.dev4670-1
        trytond_purchase_invoice_line_standalone 0.11.0.dev4647-1
        trytond_purchase_invoice_line_standalone 2.0.4.dev25079-1
        trytond_purchase_invoice_line_standalone 0.3.1.dev0-20160104
        trytond_purchase_invoice_line_standalone 0.3.0.dev0-20151222
        data_analysis_common_reader_writers_addon 0.3.0.dev0-20151217
        """.strip().splitlines()

    UNACCEPTABLE_PACKAGE_STRINGS = """
        foo a909
        barbar-242 1.2.3
        baz!bizz 0.2.5-dev1
        twobars 1.3-4-6
        .startswithadot 1.2-3
        endwithadot. 2.3-4
        .startendwithdot. 4.5.6-7
        versionstartswithoutnum .2.4.5-4
        versioneendswithadot 2.3.-2
        buildhasadot 2.3.4-2.3
        """.strip().splitlines()

    def test_pretty_package_parse(self):
        # Given
        parser = PrettyPackageStringParser(EnpkgVersion.from_string)
        package_strings = self.SCARY_PACKAGE_STRINGS

        # When
        for package_string in package_strings:
            package = parser.parse_to_package(package_string)
            expected = tuple(package_string.split())
            result = (package.name, str(package.version))

            # Then
            self.assertEqual(expected, result)

        # Given
        package_strings = self.UNACCEPTABLE_PACKAGE_STRINGS

        # When
        for package_string in package_strings:
            # Then
            with self.assertRaises(ValueError):
                parser.parse_to_package(package_string)

    def test_requirement_parsing(self):
        # Given
        package_strings = self.SCARY_PACKAGE_STRINGS

        # When
        for package_string in package_strings:
            name, version = package_string.split()
            expected = name + ' == ' + version
            result = str(Requirement._from_string(expected))

            # Then
            self.assertEqual(expected, result)

        # Given
        package_strings = self.UNACCEPTABLE_PACKAGE_STRINGS

        # When
        for package_string in package_strings:
            name, version = package_string.split()
            requirement_string = name + ' == ' + version
            # Then
            with self.assertRaises(InvalidConstraint):
                Requirement._from_string(requirement_string)



# THIS FILE IS GENERATED FROM SETUP_EXT
version = '0.8.2'
full_version = '0.8.2'
git_revision = '4ce881644c872e629358e31f37bc8d099ce42591'
is_released = True

msi_version = '0.8.2.843'

version_info = (0, 8, 2, 'final', 0)



import textwrap
import unittest

import six

from simplesat.constraints import ConstraintModifiers, Requirement
from simplesat.errors import NoPackageFound, UnexpectedlySatisfiable
from simplesat.test_utils import packages_from_definition

from ..dependency_solver import (
    minimal_unsatisfiable_subset, requirements_are_satisfiable
)


R = Requirement._from_string


class TestRequirementsAreSatistifiable(unittest.TestCase):
    def test_simple(self):
        # Given
        packages_definition = textwrap.dedent("""
        MKL 10.3-1
        MKL 11.4.1-1
        numpy 1.9.2-1; depends (MKL == 10.3-1)
        numpy 1.10.4-1; depends (MKL == 11.4.1-1)
        """)
        packages = packages_from_definition(packages_definition)

        # When/Then
        requirements = [R("numpy")]
        result = requirements_are_satisfiable(packages,  requirements)
        self.assertTrue(result.is_satisfiable)

        # When/Then
        requirements = [R("numpy < 1.10")]
        result = requirements_are_satisfiable(packages,  requirements)
        self.assertTrue(result.is_satisfiable)

        # When/Then
        requirements = [R("numpy < 1.10"), R("MKL >= 11")]
        result = requirements_are_satisfiable(packages,  requirements)
        self.assertFalse(result.is_satisfiable)

        # When/Then
        requirements = [R("numpy > 1.10"), R("MKL >= 11")]
        result = requirements_are_satisfiable(packages,  requirements)
        self.assertTrue(result.is_satisfiable)

    def test_raises_if_unresolvable_requirement(self):
        # Given
        packages_definition = textwrap.dedent("""
        MKL 11.4.1-1;
        numpy 1.10.4-1; depends (MKL == 11.4.1-1)
        """)
        packages = packages_from_definition(packages_definition)

        requirements = [R("foo")]

        # When/Then
        with self.assertRaises(NoPackageFound):
            requirements_are_satisfiable(packages,  requirements)

    def test_constraint_modifiers(self):
        # Given
        packages_definition = textwrap.dedent("""
        MKL 10.3-1
        MKL 11.4.1-1
        numpy 1.9.2-1; depends (MKL == 10.3-1)
        numpy 1.10.4-1; depends (MKL == 11.4.1-1)
        """)
        packages = packages_from_definition(packages_definition)

        requirements = [
            Requirement._from_string("numpy < 1.10"),
            Requirement._from_string("MKL >= 11")
        ]
        modifiers = ConstraintModifiers(allow_newer=("MKL",))

        # When/Then
        result = requirements_are_satisfiable(packages, requirements)
        self.assertFalse(result.is_satisfiable)
        result = requirements_are_satisfiable(packages, requirements, modifiers)
        self.assertTrue(result.is_satisfiable)


def requirements_from_definition(s):
    return [R(line) for line in s.splitlines() if line.strip()]


class TestMinUnsat(unittest.TestCase):
    def test_simple(self):
        # Given
        packages_definition = textwrap.dedent("""\
        MKL 10.3-1
        MKL 11.4.1-1
        numpy 1.9.2-1; depends (MKL == 10.3-1)
        numpy 1.10.4-1; depends (MKL == 11.4.1-1)
        """)
        packages = packages_from_definition(packages_definition)

        def callback(requirements):
            return requirements_are_satisfiable(packages, requirements).is_satisfiable

        r_min_unsat = [
            Requirement._from_string("numpy < 1.10"),
            Requirement._from_string("MKL >= 11")
        ]

        # When
        requirements = [
            Requirement._from_string("numpy < 1.10"),
            Requirement._from_string("MKL >= 11")
        ]
        min_unsat = minimal_unsatisfiable_subset(requirements, callback)

        # Then
        six.assertCountEqual(self, min_unsat, r_min_unsat)

    def test_conflicting_dependencies(self):
        # Given
        packages_definition = textwrap.dedent("""
        MKL 10.3-1
        MKL 11.4.1-1
        numpy 1.9.2-1; depends (MKL == 10.3-1)
        numpy 1.10.4-1; depends (MKL == 11.4.1-1)
        scipy 0.16.0-1; depends (MKL == 10.3-1, numpy ^= 1.9.2)
        scipy 0.17.0-1; depends (MKL == 11.4.1-1, numpy ^= 1.10.4)
        """)
        packages = packages_from_definition(packages_definition)

        def callback(requirements):
            return requirements_are_satisfiable(packages, requirements).is_satisfiable

        r_min_unsat = [
            Requirement._from_string("MKL < 11"),
            Requirement._from_string("scipy >= 0.17.0"),
        ]

        # When
        requirements = [
            Requirement._from_string("MKL < 11"),
            Requirement._from_string("numpy"),
            Requirement._from_string("scipy >= 0.17.0"),
        ]
        min_unsat = minimal_unsatisfiable_subset(requirements, callback)

        # Then
        six.assertCountEqual(self, min_unsat, r_min_unsat)

    def test_conflicting_dependencies2(self):
        # Given
        packages_definition = textwrap.dedent("""
        libgfortran 3.0.0-2
        MKL 10.3-1
        MKL 11.1.4-1
        numpy 1.7.1-1; depends (MKL == 10.3-1)
        numpy 1.9.2-3; depends (MKL == 11.1.4-1)
        pandas 0.12.0-1; depends (numpy ^= 1.7.1, python_dateutil)
        python_dateutil 2.4.2-2; depends (six ^= 1.10.0)
        pytz 2014.9.0-1
        six 1.10.0-1
        """)
        packages = packages_from_definition(packages_definition)

        requirements_definition = textwrap.dedent("""
        libgfortran == 3.0.0-2
        MKL == 11.1.4-1
        numpy == 1.9.2-3
        pandas == 0.12.0-1
        python_dateutil == 2.4.2-2
        pytz == 2014.9.0-1
        six == 1.10.0-1
        """)
        requirements = requirements_from_definition(requirements_definition)

        def callback(requirements):
            return requirements_are_satisfiable(packages, requirements).is_satisfiable

        r_min_unsat = [R("MKL == 11.1.4-1"), R("pandas == 0.12.0-1")]

        # When
        min_unsat = minimal_unsatisfiable_subset(requirements, callback)

        # Then
        six.assertCountEqual(self, min_unsat, r_min_unsat)

    def test_more_than_2_clauses(self):
        # Given
        packages_definition = textwrap.dedent("""
        A 1.0-1; provides (X)
        B 1.0-1; provides (X, Y, Z)
        C 1.0-1; provides (Y)
        D 1.0-1; provides (Y, Z)
        P 1.0-1; depends (X)
        Q 1.0-1; depends (Y); conflicts (A ^= 1.0)
        R 1.0-1; depends (Z); conflicts (B ^= 1.0)
        """)
        packages = packages_from_definition(packages_definition)

        requirements_definition = textwrap.dedent("""
        P
        Q
        R
        """)
        requirements = requirements_from_definition(requirements_definition)

        def callback(requirements):
            return requirements_are_satisfiable(packages, requirements).is_satisfiable

        r_min_unsat = [R("P"), R("Q"), R("R")]

        # When
        min_unsat = minimal_unsatisfiable_subset(requirements, callback)

        # Then
        six.assertCountEqual(self, min_unsat, r_min_unsat)

    def test_raises_unexpectedly_satisfiable(self):
        # Given
        packages_definition = textwrap.dedent("""
        MKL 11.1.4-1
        numpy 1.9.2-3; depends (MKL == 11.1.4-1)
        """)
        packages = packages_from_definition(packages_definition)

        requirements_definition = textwrap.dedent("""
        MKL == 11.1.4-1
        numpy == 1.9.2-3
        """)
        requirements = requirements_from_definition(requirements_definition)

        def callback(requirements):
            return requirements_are_satisfiable(packages, requirements)

        # When/Then
        with self.assertRaises(UnexpectedlySatisfiable):
            minimal_unsatisfiable_subset(requirements, callback)



"""
Useful code to compare solver behaviour with PHP's Composer. Obviously
don't use this in enstaller itself.
"""
import collections
import json

from ..constraints import InstallRequirement
from ..constraints.kinds import (
    Any, EnpkgUpstreamMatch, Equal, GEQ, GT, LEQ, LT
)
from ..request import JobType

# We ignore alpha/rc/etc... as composer does not allow to combine those with
# patch versions, which we use to emulate build numbers.
_TO_NORMALIZE = {
    "1.0a3": "1.0.0",
    "1.0b1": "1.0.0",
    "2010o": "2010.15.0.0",
    "2011n": "2011.14.0.0",
    "2011g": "2011.7.0.0",
    "0.14.1rc1": "0.14.1.0",
    "0.11rc1": "0.11.0.0",
    "0.9.0rc2": "0.9.0",
    "2.3b1.dev4669": "2.3.0",
    "1.2.dev213": "1.2.0",
    "0.4.2.dev2": "0.4.2",
}


JOB_KIND_TO_PHP_METHOD = {
    JobType.install: "install",
    JobType.remove: "remove",
}


def repository_to_composer_json_dict(repository):
    """ Generator of composer-compatible dict representation of an enstaller
    Repository's entries.

    Assuming the string returned by this function is written into the file
    'repository.json' file, you can load it as follows w/ Composer::

        $json_string = file_get_contents("remote.json");
        $packages = JsonFile::parseJson($json_string);

        $repository = new ArrayRepository();
        $loader = new ArrayLoader();

        foreach ($packages as $packageData) {
            $package = $loader->load($packageData):
            $repository.addPackage($package);
        }

    Parameters
    ----------
    repository : Repository
        The repository to convert
    """
    for package in repository:
        version_normalized = _normalize_php_version(package.version)
        requires = [InstallRequirement.from_constraints(p) for
                    p in package.install_requires]
        yield {
            "name": package.name,
            "version": _fix_php_version(package.version),
            "version_normalized": version_normalized,
            "require": _requirements_to_php_dict(requires),
        }


def request_to_php_parts(request):
    parts = []
    for job in request.jobs:
        name, php_constraints = _requirement_to_php_constraints(
            job.requirement
        )
        parts.append((JOB_KIND_TO_PHP_METHOD[job.kind], name, php_constraints))
    return parts


def scenario_to_php_template_variables(scenario, remote_definition,
                                       installed_definition):

    remote_repository = scenario.remote_repositories[0]
    with open(remote_definition, "wt") as fp:
        entries_iterator = repository_to_composer_json_dict(remote_repository)
        fp.write(json.dumps(list(entries_iterator), indent=4))

    installed_repository = scenario.installed_repository
    with open(installed_definition, "wt") as fp:
        entries_iterator = repository_to_composer_json_dict(
            installed_repository)
        fp.write(json.dumps(list(entries_iterator), indent=4))

    return {"remote_definition": remote_definition,
            "installed_definition": installed_definition,
            "request_parts": request_to_php_parts(scenario.request)}


def _fix_php_version(version):
    """ 'Normalize' an EnpkgVersion to a valid composer version
    string.
    """
    return str(version)


def _normalize_php_version(version):
    """ 'Normalize' an EnpkgVersion to a valid normalized composer version
    string.
    """
    upstream = str(version.upstream)
    upstream = _TO_NORMALIZE.get(upstream, upstream)
    while upstream.count(".") < 3:
        upstream += ".0"
    return "{0}-patch{1}".format(upstream, version.build)


def _normalize_php_version_constraint(version):
    """ 'Normalize' an EnpkgVersion to a valid normalized composer version
    string when used in constraints.
    """
    upstream = str(version.upstream)
    upstream = _TO_NORMALIZE.get(upstream, upstream)
    while upstream.count(".") < 3:
        upstream += ".0"

    if version.build == 0:
        return upstream
    else:
        return "{0}-patch{1}".format(upstream, version.build)


def _requirement_to_php_constraints(requirement):
    constraint_to_string = {
        GEQ: ">=",
        GT: ">",
        LEQ: "<=",
        LT: "<",
    }
    constraint_types = tuple(constraint_to_string.keys())
    parts = []
    for constraint in requirement._constraints._constraints:
        if isinstance(constraint, constraint_types):
            constraint_string = constraint_to_string[constraint.__class__]
            version = _normalize_php_version_constraint(constraint.version)
            parts.append(" ".join((constraint_string, version)))
        elif isinstance(constraint, Any):
            parts.append("*")
        else:
            raise ValueError("Unsupported constraint: %s" % constraint)

    return requirement.name, ", ".join(parts)


def _requirement_to_php_string(requirement):
    """ Convert an enstaller requirement into a composer constraint string.
    """
    parts = []
    for constraint in requirement._constraints._constraints:
        if isinstance(constraint, EnpkgUpstreamMatch):
            normalized = _normalize_php_version_constraint(constraint.version)
            parts.append("~{0}".format(normalized))
        elif isinstance(constraint, Any):
            parts.append("*")
        elif isinstance(constraint, Equal):
            normalized = _normalize_php_version_constraint(constraint.version)
            parts.append("{0}".format(normalized))
        else:
            raise NotImplementedError(constraint)
    return ", ".join(parts)


def _requirements_to_php_dict(requirements):
    """ Convert a list of requirements into a mapping
    name -> composer_requirement_string

    Parameters
    ----------
    requirements : seq
        Iterable of Requirement instances.
    """
    php_dict = collections.defaultdict(list)
    for requirement in requirements:
        php_dict[requirement.name].append(
            _requirement_to_php_string(requirement))

    return dict((k, ", ".join(v)) for k, v in php_dict.items())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import division, print_function

from collections import defaultdict, deque

import six
import itertools

from simplesat.constraints.requirement import InstallRequirement


def toposort(nodes_to_edges):
    """Return an iterator over topologically sorted groups of nodes.

    Output is a list of sets in
    topological order. The first set consists of items with no dependences,
    each subsequent set consists of items that depend upon items in the
    preceeding sets.


    Parameters
    ----------
    nodes_to_edges : dict from node to set(node)
        A directed graph expressed as a dictionary of edges whose keys are
        nodes and values are all of the nodes on which the key depends.

        For example, if node 1 depends on 2, we have ``{1: {2}, 2: set()}``.

    Yields
    ------
    set of nodes
        Each yielded set contains nodes which depend only on nodes that have
        already been yielded in a previous set. The first set contains the
        nodes with no outgoing edges.

    Raises
    ------
    ValueError
        If the graph contains cyclic dependencies.


    >>> print '\\n'.join(repr(sorted(x)) for x in toposort2({
        ...     2: set([11]),
        ...     9: set([11,8]),
        ...     10: set([11,3]),
        ...     11: set([7,5]),
        ...     8: set([7,3]),
        ...     }))
    {3, 5, 7}
    {8, 11}
    {2, 9, 10}

    """

    data = {k: v.copy() for k, v in six.iteritems(nodes_to_edges)}

    # Ignore self dependencies.
    for k, v in six.iteritems(data):
        v.discard(k)

    # Find all items that don't depend on anything.
    extra_items_in_deps = set(
        itertools.chain.from_iterable(six.itervalues(data)))
    extra_items_in_deps.difference_update(set(six.iterkeys(data)))

    # Add empty dependences where needed
    data.update({item: set() for item in extra_items_in_deps})

    while True:
        ordered = set(item for item, dep in six.iteritems(data) if not dep)
        if not ordered:
            break
        yield ordered
        data = {item: (dep - ordered)
                for item, dep in six.iteritems(data)
                if item not in ordered}
    if data:
        msg = "Cyclic dependencies exist among these items:\n{}"
        cyclic = '\n'.join(repr(x) for x in six.iteritems(data))
        raise ValueError(msg.format(cyclic))


def package_lit_dependency_graph(pool, package_lits, closed=True):
    """
    Return a dict of nodes to edges, suitable for use with :func:`toposort`.

    Parameters
    ----------
    pool : Pool
        The pool to use when resolving package literals to packages.
    package_lits : iterable of int
        The package literals to build the dependency graph for.
        These can be positive or negative. The sign will be maintained.
    closed : bool, optional
        If True, only include edges to packages dependencies that are
        themselves in `package_lits`. No package literals that are not in
        `package_lits` will appear in the graph.

    Returns
    -------
    nodes_to_edges : dict
        A dict of package_literals to sets of package_literals, as described in
        :func:`toposort`.
    """

    package_lits = tuple(package_lits)
    package_id_map = {abs(p): p for p in package_lits}
    packages = {package_id: pool.id_to_package(abs(package_id))
                for package_id in package_lits}

    R = InstallRequirement.from_constraints
    nodes_to_edges = {package_id: set() for package_id in package_lits}

    for package_lit, package in packages.items():
        for constraints in package.install_requires:
            deps = pool.what_provides(R(constraints))
            nodes_to_edges[package_lit].update(
                dep_lit for dep_lit in (
                    package_id_map.get(dep_id, dep_id)
                    for dep_id in (pool.package_id(dep) for dep in deps)
                    if (not closed or dep_id in package_id_map)
                )
            )

    return dict(nodes_to_edges)


def transitive_neighbors(nodes_to_edges):
    """ Return the set of all reachable nodes for each node in the
    nodes_to_edges adjacency dict. """
    trans = defaultdict(set)
    for node in nodes_to_edges.keys():
        _transitive(node, nodes_to_edges, trans)
    return dict(trans)


def _transitive(node, nodes_to_edges, trans):
    trans = trans if trans is not None else defaultdict(set)
    if node in trans:
        return trans
    neighbors = nodes_to_edges[node]
    trans[node].update(neighbors)
    for neighbor in neighbors:
        _transitive(neighbor, nodes_to_edges, trans)
        trans[node].update(trans[neighbor])
    return trans


def connected_nodes(node, neighbor_func, visited=None):
    """ Recursively build up a set of nodes connected to `node` by following
    neighbors as given by `neighbor_func(node)`, i.e. "flood fill."


    >>> def neighbor_func(node):
    ...     return {-node, min(node+1, 5)}
    >>> connected_nodes(0, neighbor_func)
    {-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5}
    """
    visited = set() if visited is None else visited
    queue = set([node])
    while queue:
        node = queue.pop()
        visited.add(node)
        neighbors = neighbor_func(node) - visited
        queue.update(neighbors)
    return visited


def backtrack(end, start, visited):
    """ Return a tuple of nodes from `start` to `end` by recursively looking up
    the current node in `visited`. `visited` is a dictionary of one-way edges
    between nodes.
    """
    path = [end]
    node = end
    while node != start:
        node = visited[node]
        path.append(node)
    return tuple(reversed(path))


def breadth_first_search(start, neighbor_func, targets,
                         target_func=None, visited=None):
    """
    Return an iterable of paths from `start` to each reachable terminal node
    `end`.

    Parameters
    ----------
    start : node
        The starting point of the search
    neighbor_func : callable
        Returns the neighbors of a node
    targets : set
        The nodes we're searching for. The search terminates when each member
        of `targets` has been encountered at least once, but only path is
        returned per target.
    target_func : callable, optional
        If given, then `target_func` is applied to node and the result
        is used to determine if `node` is a target via
        ``target_func(node) in targets``.
    visited : dict, optional
        If given, it will be used to track the current path. You can use it to
        directly inspect the search path after calling breadth_first_search().

    Yields
    ------
    path : tuple of nodes
        A path from node `start` to some node `end` such that
        `terminate_func(end)` is in `targets`, by following neighbors as given
        by `neighborfunc(node)`::

            >>> start = 0
            >>> targets = {10, 4}
            >>> def target_func(node):
            ...     return node*2
            >>> def neighbor_func(node):
            ...     return [node + 1]
            >>> tuple(breadth_first_search(start, neighbor_func, targets, target_func))
            ((0, 1, 2), (0, 1, 2, 3, 4, 5))
    """
    queue = deque([start])
    visited = {} if visited is None else visited
    visited[start] = None
    targets = set(targets)
    while queue:
        node = queue.popleft()
        found = target_func(node) if target_func else node
        if found in targets:
            # We found an important node. Yield the path to this node.
            targets.remove(found)
            yield backtrack(node, start, visited)
        if not targets:
            # There are no more important nodes, we're done.
            break
        for neighbor in neighbor_func(node):
            if neighbor in visited:
                continue
            queue.append(neighbor)
            visited[neighbor] = node



# -*- coding: utf-8 -*-

from .policy import DefaultPolicy
from .undetermined_clause_policy import (
    LoggedUndeterminedClausePolicy, UndeterminedClausePolicy
)

InstalledFirstPolicy = LoggedUndeterminedClausePolicy



import unittest

from simplesat.examples.van_der_waerden import van_der_waerden
from simplesat.errors import SatisfiabilityError
from ..clause import Clause
from ..minisat import MiniSATSolver


class TestMinisatProblems(unittest.TestCase):
    """Run the Minisat solver on a range of test problems.
    """
    def test_unit_assignments_are_remembered(self):
        # Regression test for #31 (calling _setup_assignments() after
        # enqueueing unit clauses caused unit assignments to be erased).

        # Given
        s = MiniSATSolver()
        s.add_clause([1, 2, 3])
        s.add_clause([-1, 2, 3])
        s.add_clause([-2])
        s.add_clause([-3])
        s._setup_assignments()

        # Then
        with self.assertRaises(SatisfiabilityError):
            s.search()

    def test_no_assumptions(self):
        # A simple problem with only unit propagation, no assumptions (except
        # for the initial one), no backtracking, and no conflicts.

        # Given
        s = MiniSATSolver()
        cl1 = Clause([1, -2])
        cl2 = Clause([1,  2, -3])
        cl3 = Clause([1,  2,  3, -4])
        s.add_clause(cl1)
        s.add_clause(cl2)
        s.add_clause(cl3)
        s._setup_assignments()

        # When
        sol = s.search()

        # Then
        result = dict(sol.items())
        self.assertEqual(result, {1: True, 2: True, 3: True, 4: True})

    def test_one_assumption(self):
        # A simple problem with only unit propagation, where one additional
        # assumption needs to be made.

        # Given
        s = MiniSATSolver()
        cl1 = Clause([1, -2])
        cl3 = Clause([-1,  -2,  3, -4])
        s.add_clause(cl1)
        s.add_clause(cl3)
        s._setup_assignments()

        # When
        sol = s.search()

        # Then
        result = dict(sol.items())
        self.assertEqual(result, {1: True, 2: True, 3: True, 4: True})


def check_solution(clauses, solution):
    for clause in clauses:
        for lit in clause:
            if solution.value(lit):
                # Clause is satisfied.
                break
        else:
            # No true literal
            return False
    return True


class TestMinisatVanDerWaerden(unittest.TestCase):

    def test_van_der_waerden_solvable(self):
        # Given
        j, k, n = 3, 3, 8
        s = MiniSATSolver()
        clauses = van_der_waerden(j, k, n)
        for clause in clauses:
            s.add_clause(clause)
        s._setup_assignments()

        # When
        solution = s.search()

        # Then
        self.assertTrue(check_solution(s.clauses, solution),
                        msg='{} does not satisfy SAT problem'.format(solution))

    def test_van_der_waerden_not_solvable(self):
        # Given
        j, k, n = 3, 3, 9
        s = MiniSATSolver()
        clauses = van_der_waerden(j, k, n)
        for clause in clauses:
            s.add_clause(clause)
        s._setup_assignments()

        # Then
        with self.assertRaises(SatisfiabilityError):
            s.search()



from __future__ import absolute_import

import copy
import sys
import unittest

from six.moves import cPickle

from .. import DefaultOrderedDict


class TestDefaultOrderedDict(unittest.TestCase):
    def test_simple(self):
        # Given
        data = DefaultOrderedDict(list)

        # When
        data[1].append(1)
        data[0].append(0)

        # Then
        self.assertEqual(list(data.keys()), [1, 0])
        self.assertEqual(data[0], [0])
        self.assertEqual(data[1], [1])
        self.assertEqual(data[2], [])
        if sys.version_info[0] == 2:
            r_repr = ("DefaultOrderedDict(<type 'list'>, "
                      "DefaultOrderedDict([(1, [1]), (0, [0]), (2, [])]))")
        else:
            r_repr = ("DefaultOrderedDict(<class 'list'>, "
                      "DefaultOrderedDict([(1, [1]), (0, [0]), (2, [])]))")
        self.assertEqual(repr(data), r_repr)

    def test_pickling(self):
        # Given
        data = DefaultOrderedDict(list)
        data[1].append(1)
        data[0].append(0)

        # When
        s = cPickle.dumps(data)
        unpickled_data = cPickle.loads(s)

        # Then
        self.assertEqual(unpickled_data, data)
        self.assertIsInstance(unpickled_data, (DefaultOrderedDict,))

    def test_copy(self):
        # Given
        data = DefaultOrderedDict(list)
        data[1].append(1)

        # When
        data_copy = data.copy()

        # Then
        self.assertEqual(data, data_copy)
        self.assertIsInstance(data_copy, DefaultOrderedDict)

        # When
        # Check we don't do deep copy
        data[1].append(2)

        # Then
        self.assertEqual(data, data_copy)

        # When
        data.pop(1)
        self.assertEqual(data_copy[1], [1, 2])
        self.assertIsInstance(data_copy, (DefaultOrderedDict,))

    def test_deepcopy(self):
        # Given
        data = DefaultOrderedDict(list)
        data[1].append(1)

        # When
        data_copy = copy.deepcopy(data)

        # Then
        self.assertEqual(data, data_copy)
        self.assertIsInstance(data_copy, (DefaultOrderedDict,))

        # When
        data[1].append(2)

        # Then
        self.assertNotEqual(data_copy[1], data[1])



from okonomiyaki.versions import EnpkgVersion

from .parser import _RawConstraintsParser


class MultiConstraints(object):
    """
    A set of constraints to match a version against.

    Example
    -------

    >>> V = EnpkgVersion.from_string
    >>> m = MultiConstraints([GEQ(V("1.3.0-1"))])
    >>> m.matches(V("1.3.0-1"))
    True
    >>> m.matches(V("1.2.0-1"))
    False
    """
    @classmethod
    def _from_string(cls, requirement_string,
                     version_factory=EnpkgVersion.from_string):
        """Creates a MultiConstraints from a requirement string.

        Mostly useful to write readable tests, do not use in actual
        code.
        """
        parser = _RawConstraintsParser()
        return cls(parser.parse(requirement_string, version_factory))

    def __init__(self, constraints=None):
        if constraints is None:
            self._constraints = tuple()
        else:
            self._constraints = tuple(constraints)

    def matches(self, version_candidate):
        """ Returns True if the given version matches this set of
        requirements.

        Parameters
        ----------
        version_candidate : Version
            A comparable version instance. Must match the version class
            used for the set of requirements.
        """
        for constraint in self._constraints:
            if not constraint.matches(version_candidate):
                return False
        return True

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self._constraints == other._constraints)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self._constraints)
